# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
import sklearn.tree

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "A decision tree regressor.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "criterion",
                "splitter",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "max_features",
            ],
            "relevantToOptimizer": [
                "criterion",
                "splitter",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "max_features",
            ],
            "additionalProperties": False,
            "properties": {
                "criterion": {
                    "description": "Function to measure the quality of a split.",
                    "anyOf": [
                        {"enum": ["mse", "friedman_mse"]},
                        {"enum": ["mae"], "forOptimizer": False},
                    ],
                    "default": "mse",
                },
                "splitter": {
                    "enum": ["best", "random"],
                    "default": "best",
                    "description": "Strategy to choose the split at each node.",
                },
                "max_depth": {
                    "description": "Maximum depth of the tree.",
                    "default": None,
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 5,
                        },
                        {
                            "enum": [None],
                            "description": "If None, then nodes are expanded until all leaves are pure, or until all leaves contain less than min_samples_split samples.",
                        },
                    ],
                },
                "min_samples_split": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 2,
                            "laleMaximum": "X/maxItems",  # number of rows
                            "forOptimizer": False,
                            "description": "Consider min_samples_split as the minimum number.",
                        },
                        {
                            "type": "number",
                            "minimum": 0.0,
                            "exclusiveMinimum": True,
                            "maximum": 1.0,
                            "minimumForOptimizer": 0.01,
                            "maximumForOptimizer": 0.5,
                            "default": 0.05,
                            "description": "min_samples_split is a fraction and ceil(min_samples_split * n_samples) are the minimum number of samples for each split.",
                        },
                    ],
                    "default": 2,
                    "description": "The minimum number of samples required to split an internal node.",
                },
                "min_samples_leaf": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "laleMaximum": "X/maxItems",  # number of rows
                            "forOptimizer": False,
                            "description": "Consider min_samples_leaf as the minimum number.",
                        },
                        {
                            "type": "number",
                            "minimum": 0.0,
                            "exclusiveMinimum": True,
                            "maximum": 0.5,
                            "minimumForOptimizer": 0.01,
                            "default": 0.05,
                            "description": "min_samples_leaf is a fraction and ceil(min_samples_leaf * n_samples) are the minimum number of samples for each node.",
                        },
                    ],
                    "default": 1,
                    "description": "The minimum number of samples required to be at a leaf node.",
                },
                "min_weight_fraction_leaf": {
                    "type": "number",
                    "minimum": 0.0,
                    "maximum": 0.5,
                    "default": 0.0,
                    "description": "Minimum weighted fraction of the sum total of weights (of all the input samples) required to be at a leaf node.",
                },
                "max_features": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 2,
                            "laleMaximum": "X/items/maxItems",  # number of columns
                            "forOptimizer": False,
                            "description": "Consider max_features features at each split.",
                        },
                        {
                            "type": "number",
                            "minimum": 0.0,
                            "exclusiveMinimum": True,
                            "maximum": 1.0,
                            "minimumForOptimizer": 0.01,
                            "default": 0.5,
                            "distribution": "uniform",
                            "description": "max_features is a fraction and int(max_features * n_features) features are considered at each split.",
                        },
                        {"enum": ["auto", "sqrt", "log2", None]},
                    ],
                    "default": "auto",
                    "description": "The number of features to consider when looking for the best split.",
                },
                "random_state": {
                    "description": "Seed of pseudo-random number generator.",
                    "anyOf": [
                        {"laleType": "numpy.random.RandomState"},
                        {
                            "description": "RandomState used by np.random",
                            "enum": [None],
                        },
                        {"description": "Explicit seed.", "type": "integer"},
                    ],
                    "default": None,
                },
                "max_leaf_nodes": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 1000,
                        },
                        {
                            "enum": [None],
                            "description": "Unlimited number of leaf nodes.",
                        },
                    ],
                    "default": None,
                    "description": "Grow a tree with ``max_leaf_nodes`` in best-first fashion.",
                },
                "min_impurity_decrease": {
                    "type": "number",
                    "default": 0.0,
                    "minimum": 0.0,
                    "maximumForOptimizer": 10.0,
                    "description": "A node will be split if this split induces a decrease of the impurity greater than or equal to this value.",
                },
                "min_impurity_split": {
                    "anyOf": [{"type": "number", "minimum": 0.0}, {"enum": [None]}],
                    "default": None,
                    "description": "Threshold for early stopping in tree growth.",
                },
                "presort": {
                    "type": "boolean",
                    "default": False,
                    "description": "Whether to presort the data to speed up the finding of best splits in fitting.",
                },
            },
        }
    ],
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "The target values (real numbers).",
        },
        "sample_weight": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"enum": [None], "description": "Samples are equally weighted."},
            ],
            "description": "Sample weights.",
        },
        "check_input": {
            "type": "boolean",
            "default": True,
            "description": "Allow to bypass several input checking.",
        },
        "X_idx_sorted": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
                {"enum": [None]},
            ],
            "default": None,
            "description": "The indexes of the sorted training input samples. If many tree",
        },
    },
}

_input_predict_schema = {
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "check_input": {
            "type": "boolean",
            "default": True,
            "description": "Allow to bypass several input checking.",
        },
    },
}

_output_predict_schema = {
    "description": "The predicted classes, or the predict values.",
    "type": "array",
    "items": {"type": "number"},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Decision tree regressor`_ from scikit-learn.

.. _`Decision tree regressor`: https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.decision_tree_regressor.html",
    "import_from": "sklearn.tree",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}


DecisionTreeRegressor: lale.operators.PlannedIndividualOp
DecisionTreeRegressor = lale.operators.make_operator(
    sklearn.tree.DecisionTreeRegressor, _combined_schemas
)

if sklearn.__version__ >= "0.22":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.tree.DecisionTreeRegressor.html
    # new: https://scikit-learn.org/0.22/modules/generated/sklearn.tree.DecisionTreeRegressor.html
    from lale.schemas import AnyOf, Bool, Enum, Float

    DecisionTreeRegressor = DecisionTreeRegressor.customize_schema(
        presort=AnyOf(
            types=[Bool(), Enum(["deprecated"])],
            desc="This parameter is deprecated and will be removed in v0.24.",
            default="deprecated",
        ),
        ccp_alpha=Float(
            desc="Complexity parameter used for Minimal Cost-Complexity Pruning. The subtree with the largest cost complexity that is smaller than ccp_alpha will be chosen. By default, no pruning is performed.",
            default=0.0,
            forOptimizer=False,
            minimum=0.0,
            maximumForOptimizer=0.1,
        ),
        set_as_available=True,
    )

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.22/modules/generated/sklearn.tree.DecisionTreeRegressor.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.tree.DecisionTreeRegressor.html
    DecisionTreeRegressor = DecisionTreeRegressor.customize_schema(
        criterion={
            "description": "Function to measure the quality of a split.",
            "anyOf": [
                {"enum": ["mse", "friedman_mse", "poisson"]},
                {"enum": ["mae"], "forOptimizer": False},
            ],
            "default": "mse",
        },
        presort=None,
        set_as_available=True,
    )

if sklearn.__version__ >= "1.0":
    # old: https://scikit-learn.org/0.24/modules/generated/sklearn.tree.DecisionTreeRegressor.html
    # new: https://scikit-learn.org/1.0/modules/generated/sklearn.tree.DecisionTreeRegressor.html
    DecisionTreeRegressor = DecisionTreeRegressor.customize_schema(
        criterion={
            "description": "Function to measure the quality of a split.",
            "anyOf": [
                {
                    "enum": [
                        "squared_error",
                        "friedman_mse",
                        "absolute_error",
                        "poisson",
                    ]
                },
                {"enum": ["mae", "mse"], "forOptimizer": False},
            ],
            "default": "squared_error",
        },
        min_impurity_split=None,
        set_as_available=True,
    )

lale.docstrings.set_docstrings(DecisionTreeRegressor)



# Copyright 2019,2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
try:
    import snapml  # type: ignore

    snapml_installed = True
except ImportError:
    snapml_installed = False

import lale.datasets.data_schemas
import lale.docstrings
import lale.operators


class _SnapLogisticRegressionImpl:
    def __init__(
        self,
        max_iter=1000,
        regularizer=1.0,
        device_ids=None,
        verbose=False,
        use_gpu=False,
        class_weight=None,
        dual=True,
        n_jobs=1,
        penalty="l2",
        tol=0.001,
        generate_training_history=None,
        privacy=False,
        eta=0.3,
        batch_size=100,
        privacy_epsilon=10,
        grad_clip=1,
        fit_intercept=False,
        intercept_scaling=1.0,
        normalize=False,
        kernel=None,
        gamma=1.0,
        n_components=100,
        random_state=None,
    ):

        assert (
            snapml_installed
        ), """Your Python environment does not have snapml installed. Install using: pip install snapml"""
        self._hyperparams = {
            "max_iter": max_iter,
            "regularizer": regularizer,
            "device_ids": device_ids,
            "verbose": verbose,
            "use_gpu": use_gpu,
            "class_weight": class_weight,
            "dual": dual,
            "n_jobs": n_jobs,
            "penalty": penalty,
            "tol": tol,
            "generate_training_history": generate_training_history,
            "privacy": privacy,
            "eta": eta,
            "batch_size": batch_size,
            "privacy_epsilon": privacy_epsilon,
            "grad_clip": grad_clip,
            "fit_intercept": fit_intercept,
            "intercept_scaling": intercept_scaling,
            "normalize": normalize,
            "kernel": kernel,
            "gamma": gamma,
            "n_components": n_components,
            "random_state": random_state,
        }
        modified_hps = {**self._hyperparams}
        if modified_hps["device_ids"] is None:
            modified_hps["device_ids"] = [0]  # TODO: support list as default
        self._wrapped_model = snapml.SnapLogisticRegression(**modified_hps)

    def fit(self, X, y, **fit_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        y = lale.datasets.data_schemas.strip_schema(y)
        self._wrapped_model.fit(X, y, **fit_params)
        return self

    def predict(self, X, **predict_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        return self._wrapped_model.predict(X, **predict_params)

    def predict_proba(self, X, **predict_proba_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        return self._wrapped_model.predict_proba(X, **predict_proba_params)


_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "relevantToOptimizer": [
                "fit_intercept",
                "regularizer",
                "max_iter",
            ],
            "additionalProperties": False,
            "properties": {
                "max_iter": {
                    "type": "integer",
                    "minimum": 1,
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "default": 100,
                    "description": "Maximum number of iterations used by the solver to converge.",
                },
                "regularizer": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 1.0,
                    "exclusiveMinimum": True,
                    "minimumForOptimizer": 1.0,
                    "maximumForOptimizer": 100.0,
                    "distribution": "uniform",
                    "description": "Larger regularization values imply stronger regularization.",
                },
                "use_gpu": {
                    "type": "boolean",
                    "default": False,
                    "description": "Use GPU Acceleration.",
                },
                "device_ids": {
                    "anyOf": [
                        {"description": "Use [0].", "enum": [None]},
                        {"type": "array", "items": {"type": "integer"}},
                    ],
                    "default": None,
                    "description": "Device IDs of the GPUs which will be used when GPU acceleration is enabled.",
                },
                "class_weight": {
                    "enum": ["balanced", None],
                    "default": None,
                    "description": "If set to 'balanced' samples weights will be applied to account for class imbalance, otherwise no sample weights will be used.",
                },
                "dual": {
                    "type": "boolean",
                    "default": True,
                    "description": "Use dual formulation (rather than primal).",
                },
                "verbose": {
                    "type": "boolean",
                    "default": False,
                    "description": "If True, it prints the training cost, one per iteration. Warning: this will increase the training time. For performance evaluation, use verbose=False.",
                },
                "n_jobs": {
                    "type": "integer",
                    "minimum": 1,
                    "default": 1,
                    "description": "The number of threads used for running the training. The value of this parameter should be a multiple of 32 if the training is performed on GPU (use_gpu=True).",
                },
                "penalty": {
                    "enum": ["l1", "l2"],
                    "default": "l2",
                    "description": "The regularization / penalty type. Possible values are 'l2' for L2 regularization (LogisticRegression) or 'l1' for L1 regularization (SparseLogisticRegression). L1 regularization is possible only for the primal optimization problem (dual=False).",
                },
                "tol": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 0.001,
                    "exclusiveMinimum": True,
                    "description": "The tolerance parameter. Training will finish when maximum change in model coefficients is less than tol.",
                },
                "generate_training_history": {
                    "enum": ["summary", "full", None],
                    "default": None,
                    "description": "Determines the level of summary statistics that are generated during training.",
                },
                "privacy": {
                    "type": "boolean",
                    "default": False,
                    "description": "Train the model using a differentially private algorithm.",
                },
                "eta": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 0.3,
                    "exclusiveMinimum": True,
                    "description": "Learning rate for the differentially private training algorithm.",
                },
                "batch_size": {
                    "type": "integer",
                    "minimum": 1,
                    "default": 100,
                    "description": "Mini-batch size for the differentially private training algorithm.",
                },
                "privacy_epsilon": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 10.0,
                    "exclusiveMinimum": True,
                    "description": "Target privacy gaurantee. Learned model will be (privacy_epsilon, 0.01)-private.",
                },
                "grad_clip": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 1.0,
                    "description": "Gradient clipping parameter for the differentially private training algorithm.",
                },
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "Add bias term -- note, may affect speed of convergence, especially for sparse datasets.",
                },
                "intercept_scaling": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 1.0,
                    "exclusiveMinimum": True,
                    "description": "Scaling of bias term. The inclusion of a bias term is implemented by appending an additional feature to the dataset. This feature has a constant value, that can be set using this parameter.",
                },
                "normalize": {
                    "type": "boolean",
                    "default": True,
                    "description": "Normalize rows of dataset (recommended for fast convergence).",
                },
                "kernel": {
                    "enum": ["rbf", "linear"],
                    "default": "linear",
                    "description": "Approximate feature map of a specified kernel function.",
                },
                "gamma": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 1.0,
                    "exclusiveMinimum": True,
                    "description": "Parameter of RBF kernel: exp(-gamma * x^2).",
                },
                "n_components": {
                    "type": "integer",
                    "minimum": 1,
                    "default": 100,
                    "description": "Dimensionality of the feature space when approximating a kernel function.",
                },
                "random_state": {
                    "description": "Seed of pseudo-random number generator.",
                    "anyOf": [
                        {
                            "description": "RandomState used by np.random",
                            "enum": [None],
                        },
                        {"description": "Explicit seed.", "type": "integer"},
                    ],
                    "default": None,
                },
            },
        },
        {
            "description": "L1 regularization is supported only for primal optimization problems.",
            "anyOf": [
                {"type": "object", "properties": {"penalty": {"enum": ["l2"]}}},
                {"type": "object", "properties": {"dual": {"enum": [False]}}},
            ],
        },
        {
            "description": "Privacy only supported for primal objective functions.",
            "anyOf": [
                {"type": "object", "properties": {"privacy": {"enum": [False]}}},
                {"type": "object", "properties": {"dual": {"enum": [False]}}},
            ],
        },
        {
            "description": "Privacy only supported for L2-regularized objective functions.",
            "anyOf": [
                {"type": "object", "properties": {"privacy": {"enum": [False]}}},
                {"type": "object", "properties": {"penalty": {"enum": ["l2"]}}},
            ],
        },
        {
            "description": "Privacy not supported with fit_intercept=True.",
            "anyOf": [
                {"type": "object", "properties": {"privacy": {"enum": [False]}}},
                {"type": "object", "properties": {"fit_intercept": {"enum": [False]}}},
            ],
        },
    ],
}

_input_fit_schema = {
    "description": "Fit the model according to the given train dataset.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "y": {
            "description": "The classes.",
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"type": "array", "items": {"type": "string"}},
                {"type": "array", "items": {"type": "boolean"}},
            ],
        },
    },
}

_input_predict_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "n_jobs": {
            "type": "integer",
            "minimum": 0,
            "default": 0,
            "description": "Number of threads used to run inference. By default inference runs with maximum number of available threads.",
        },
    },
}

_output_predict_schema = {
    "description": "The predicted classes.",
    "anyOf": [
        {"type": "array", "items": {"type": "number"}},
        {"type": "array", "items": {"type": "string"}},
        {"type": "array", "items": {"type": "boolean"}},
    ],
}

_input_predict_proba_schema = {
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "n_jobs": {
            "type": "integer",
            "minimum": 0,
            "default": 0,
            "description": "Number of threads used to run inference. By default inference runs with maximum number of available threads.",
        },
    },
}

_output_predict_proba_schema = {
    "type": "array",
    "description": "The outer array is over samples aka rows.",
    "items": {
        "type": "array",
        "description": "The inner array contains probabilities corresponding to each class.",
        "items": {"type": "number"},
    },
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Logistic Regression`_ from `Snap ML`_.

.. _`Logistic Regression`: https://snapml.readthedocs.io/en/latest/#snapml.LogisticRegression
.. _`Snap ML`: https://www.zurich.ibm.com/snapml/
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.snapml.snap_logistic_regression.html",
    "import_from": "snapml",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}


SnapLogisticRegression = lale.operators.make_operator(
    _SnapLogisticRegressionImpl, _combined_schemas
)

lale.docstrings.set_docstrings(SnapLogisticRegression)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import autoai_libs.cognito.transforms.transform_utils

import lale.docstrings
import lale.operators


class _TA2Impl:
    def __init__(
        self,
        fun,
        name,
        datatypes1,
        feat_constraints1,
        datatypes2,
        feat_constraints2,
        tgraph=None,
        apply_all=True,
        col_names=None,
        col_dtypes=None,
        col_as_json_objects=None,
    ):
        self._hyperparams = {
            "fun": fun,
            "name": name,
            "datatypes1": datatypes1,
            "feat_constraints1": feat_constraints1,
            "datatypes2": datatypes2,
            "feat_constraints2": feat_constraints2,
            "tgraph": tgraph,
            "apply_all": apply_all,
            "col_names": col_names,
            "col_dtypes": col_dtypes,
            "col_as_json_objects": col_as_json_objects,
        }
        self._wrapped_model = autoai_libs.cognito.transforms.transform_utils.TA2(
            **self._hyperparams
        )

    def fit(self, X, y=None, **fit_params):
        self._wrapped_model.fit(X, y, **fit_params)
        return self

    def transform(self, X):
        result = self._wrapped_model.transform(X)
        return result


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "fun",
                "name",
                "datatypes1",
                "feat_constraints1",
                "datatypes2",
                "feat_constraints2",
                "tgraph",
                "apply_all",
                "col_names",
                "col_dtypes",
                "col_as_json_objects",
            ],
            "relevantToOptimizer": [],
            "properties": {
                "fun": {
                    "description": "The function pointer.",
                    "laleType": "Any",
                    "default": None,
                },
                "name": {
                    "description": "A string name that uniquely identifies this transformer from others.",
                    "anyOf": [{"type": "string"}, {"enum": [None]}],
                    "default": None,
                },
                "datatypes1": {
                    "description": "List of datatypes that are valid input to the first argument of the transformer function (`numeric`, `float`, `int`, `integer`).",
                    "anyOf": [
                        {"type": "array", "items": {"type": "string"}},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "feat_constraints1": {
                    "description": "All constraints that must be satisfied by a column to be considered a valid first argument to this transform.",
                    "laleType": "Any",
                    "default": None,
                },
                "datatypes2": {
                    "description": "List of datatypes that are valid input to the second argument of the transformer function (numeric, float, int, etc.).",
                    "anyOf": [
                        {"type": "array", "items": {"type": "string"}},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "feat_constraints2": {
                    "description": "All constraints that must be satisfied by a column to be considered a valid second argument to this transform.",
                    "laleType": "Any",
                    "default": None,
                },
                "tgraph": {
                    "description": "Should be the invoking TGraph() object.",
                    "anyOf": [
                        {"laleType": "Any"},
                        {
                            "enum": [None],
                            "description": "Passing None will result in some failure to detect some inefficiencies due to lack of caching.",
                        },
                    ],
                    "default": None,
                },
                "apply_all": {
                    "description": "Only use applyAll = True. It means that the transformer will enumerate all features (or feature sets) that match the specified criteria and apply the provided function to each.",
                    "type": "boolean",
                    "default": True,
                },
                "col_names": {
                    "description": "Names of the feature columns in a list.",
                    "anyOf": [
                        {"type": "array", "items": {"type": "string"}},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "col_dtypes": {
                    "description": "List of the datatypes of the feature columns.",
                    "anyOf": [
                        {"type": "array", "items": {"laleType": "Any"}},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "col_as_json_objects": {
                    "description": "Names of the feature columns in a json dict.",
                    "anyOf": [{"type": "object"}, {"enum": [None]}],
                    "default": None,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"laleType": "Any"}},
        },
        "y": {"laleType": "Any"},
    },
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"laleType": "Any"}}}
    },
}

_output_transform_schema = {
    "description": "Features; the outer array is over samples.",
    "type": "array",
    "items": {"type": "array", "items": {"laleType": "Any"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Operator from `autoai_libs`_. Feature transformation for binary stateless functions, such as sum or product.

.. _`autoai_libs`: https://pypi.org/project/autoai-libs""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_libs.ta2.html",
    "import_from": "autoai_libs.cognito.transforms.transform_utils",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


TA2 = lale.operators.make_operator(_TA2Impl, _combined_schemas)

lale.docstrings.set_docstrings(TA2)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
from sklearn.linear_model import SGDClassifier as SKLModel

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "inherited docstring for SGDClassifier Linear classifiers (SVM, logistic regression, a.o.) with SGD training.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "loss",
                "penalty",
                "alpha",
                "l1_ratio",
                "fit_intercept",
                "max_iter",
                "tol",
                "shuffle",
                "verbose",
                "epsilon",
                "n_jobs",
                "random_state",
                "learning_rate",
                "eta0",
                "power_t",
                "early_stopping",
                "validation_fraction",
                "n_iter_no_change",
                "class_weight",
                "warm_start",
                "average",
            ],
            "relevantToOptimizer": [
                "loss",
                "penalty",
                "alpha",
                "l1_ratio",
                "fit_intercept",
                "max_iter",
                "tol",
                "shuffle",
                "epsilon",
                "learning_rate",
                "eta0",
                "power_t",
            ],
            "additionalProperties": False,
            "properties": {
                "loss": {
                    "enum": [
                        "hinge",
                        "log",
                        "modified_huber",
                        "squared_hinge",
                        "perceptron",
                        "squared_loss",
                        "huber",
                        "epsilon_insensitive",
                        "squared_epsilon_insensitive",
                    ],
                    "default": "hinge",
                    "description": "The loss function to be used. Defaults to 'hinge', which gives a linear SVM.",
                },
                "penalty": {
                    "description": "The penalty (aka regularization term) to be used. Defaults to 'l2'",
                    "enum": ["elasticnet", "l1", "l2"],
                    "default": "l2",
                },
                "alpha": {
                    "type": "number",
                    "minimumForOptimizer": 1e-10,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "Constant that multiplies the regularization term. Defaults to 0.0001",
                },
                "l1_ratio": {
                    "type": "number",
                    "minimumForOptimizer": 1e-9,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 0.15,
                    "description": "The Elastic Net mixing parameter, with 0 <= l1_ratio <= 1.",
                },
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether the intercept should be estimated or not. If False, the",
                },
                "max_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 1000,
                    "description": "The maximum number of passes over the training data (aka epochs).",
                },
                "tol": {
                    "anyOf": [
                        {
                            "type": "number",
                            "minimumForOptimizer": 1e-08,
                            "maximumForOptimizer": 0.01,
                            "distribution": "loguniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": 0.001,
                    "description": "The stopping criterion.",
                },
                "shuffle": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether or not the training data should be shuffled after each epoch.",
                },
                "verbose": {
                    "type": "integer",
                    "default": 0,
                    "description": "The verbosity level",
                },
                "epsilon": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 1.35,
                    "distribution": "loguniform",
                    "default": 0.1,
                    "description": "Epsilon in the epsilon-insensitive loss functions; only if `loss` is",
                },
                "n_jobs": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": None,
                    "description": "The number of CPUs to use to do the OVA (One Versus All, for",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "The seed of the pseudo random number generator to use when shuffling",
                },
                "learning_rate": {
                    "enum": ["optimal", "constant", "invscaling", "adaptive"],
                    "default": "optimal",
                    "description": "The learning rate schedule:",
                },
                "eta0": {
                    "type": "number",
                    "minimumForOptimizer": 0.01,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 0.0,
                    "description": "The initial learning rate for the 'constant', 'invscaling' or",
                },
                "power_t": {
                    "type": "number",
                    "minimumForOptimizer": 0.00001,
                    "maximumForOptimizer": 1.0,
                    "distribution": "uniform",
                    "default": 0.5,
                    "description": "The exponent for inverse scaling learning rate [default 0.5].",
                },
                "early_stopping": {
                    "type": "boolean",
                    "default": False,
                    "description": "Whether to use early stopping to terminate training when validation",
                },
                "validation_fraction": {
                    "type": "number",
                    "minimum": 0.0,
                    "maximum": 1.0,
                    "default": 0.1,
                    "description": "The proportion of training data to set aside as validation set for",
                },
                "n_iter_no_change": {
                    "type": "integer",
                    "minimumForOptimizer": 5,
                    "maximumForOptimizer": 10,
                    "default": 5,
                    "description": "Number of iterations with no improvement to wait before early stopping.",
                },
                "class_weight": {
                    "anyOf": [{"type": "object"}, {"enum": ["balanced", None]}],
                    "default": None,
                    "description": "Preset for the class_weight fit parameter.",
                },
                "warm_start": {
                    "type": "boolean",
                    "default": False,
                    "description": "When set to True, reuse the solution of the previous call to fit as",
                },
                "average": {
                    "anyOf": [
                        {"type": "boolean"},
                        {"type": "integer", "forOptimizer": False},
                    ],
                    "default": False,
                    "description": "When set to True, computes the averaged SGD weights and stores the result in the ``coef_`` attribute.",
                },
            },
        },
        {
            "description": "eta0 must be greater than 0 if the learning_rate is not ‘optimal’.",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {
                        "learning_rate": {"enum": ["optimal"]},
                    },
                },
                {
                    "type": "object",
                    "properties": {
                        "eta0": {
                            "type": "number",
                            "minimum": 0.0,
                            "exclusiveMinimum": True,
                        },
                    },
                },
            ],
        },
    ],
}

_input_fit_schema = {
    "description": "Fit linear model with Stochastic Gradient Descent.",
    "required": ["X", "y"],
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "Training data",
        },
        "y": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"type": "array", "items": {"type": "string"}},
                {"type": "array", "items": {"type": "boolean"}},
            ],
            "description": "Target values",
        },
        "coef_init": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The initial coefficients to warm-start the optimization.",
        },
        "intercept_init": {
            "type": "array",
            "items": {"type": "number"},
            "description": "The initial intercept to warm-start the optimization.",
        },
        "sample_weight": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "number"},
                },
                {"enum": [None]},
            ],
            "default": None,
            "description": "Weights applied to individual samples.",
        },
    },
}
_input_predict_schema = {
    "description": "Predict class labels for samples in X.",
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "Training data",
        },
    },
}
_output_predict_schema = {
    "description": "Predicted class label per sample.",
    "anyOf": [
        {"type": "array", "items": {"type": "number"}},
        {"type": "array", "items": {"type": "string"}},
        {"type": "array", "items": {"type": "boolean"}},
    ],
}

_input_predict_proba_schema = {
    "description": "Probability estimates.",
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
        },
    },
}
_output_predict_proba_schema = {
    "description": "Returns the probability of the sample for each class in the model,",
    "type": "array",
    "items": {
        "type": "array",
        "items": {"type": "number"},
    },
}

_input_decision_function_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        }
    },
}

_output_decision_function_schema = {
    "description": "Confidence scores for samples for each class in the model.",
    "anyOf": [
        {
            "description": "In the multi-way case, score per (sample, class) combination.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        {
            "description": "In the binary case, score for `self._classes[1]`.",
            "type": "array",
            "items": {"type": "number"},
        },
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`SGD classifier`_ from scikit-learn uses linear classifiers (SVM, logistic regression, a.o.) with stochastic gradient descent training.

.. _`SGD classifier`: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDClassifier.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.sgd_classifier.html",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
        "input_decision_function": _input_decision_function_schema,
        "output_decision_function": _output_decision_function_schema,
    },
}


SGDClassifier = lale.operators.make_operator(SKLModel, _combined_schemas)

if sklearn.__version__ >= "1.0":
    # old: https://scikit-learn.org/0.24/modules/generated/sklearn.linear_model.SGDClassifer.html
    # new: https://scikit-learn.org/1.0/modules/generated/sklearn.linear_model.SGDClassifier.html
    SGDClassifier = SGDClassifier.customize_schema(
        loss={
            "description": """The loss function to be used. Defaults to ‘hinge’, which gives a linear SVM.
The possible options are ‘hinge’, ‘log’, ‘modified_huber’, ‘squared_hinge’, ‘perceptron’,
or a regression loss: ‘squared_error’, ‘huber’, ‘epsilon_insensitive’, or ‘squared_epsilon_insensitive’.
The ‘log’ loss gives logistic regression, a probabilistic classifier.
‘modified_huber’ is another smooth loss that brings tolerance to outliers as well as probability estimates.
‘squared_hinge’ is like hinge but is quadratically penalized.
‘perceptron’ is the linear loss used by the perceptron algorithm.
The other losses are designed for regression but can be useful in classification as well; see SGDRegressor for a description.
More details about the losses formulas can be found in the scikit-learn User Guide.""",
            "anyOf": [
                {
                    "enum": [
                        "hinge",
                        "log",
                        "modified_huber",
                        "squared_hinge",
                        "perceptron",
                        "squared_error",
                        "huber",
                        "epsilon_insensitive",
                        "squared_epsilon_insensitive",
                    ],
                },
                {"enum": ["squared_loss"], "forOptimizer": False},
            ],
            "default": "hinge",
        },
        set_as_available=True,
    )

lale.docstrings.set_docstrings(SGDClassifier)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional

from ConfigSpace.conditions import EqualsCondition
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Hyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
from smac.configspace import ConfigurationSpace

from lale.search.PGO import PGO
from lale.search.search_space import (
    SearchSpace,
    SearchSpaceArray,
    SearchSpaceEnum,
    SearchSpaceNumber,
    should_print_search_space,
)
from lale.search.search_space_grid import SearchSpaceGrid, get_search_space_grids

if TYPE_CHECKING:
    import lale.operators as Ops


def lale_op_smac_tae(op: "Ops.PlannedOperator", f_min):
    # TODO: we can probably do this in a different way, but get_smac_configuration_space
    # we already have these sklearn compatibility wrappers it is easier for now to use them
    op_compat = op

    def f(cfg):
        from sklearn.base import clone

        wrapped_op = clone(op_compat)
        cfg2 = smac_fixup_params(cfg)
        trainable = wrapped_op.set_params(**cfg2)

        return f_min(trainable)

    return f


def lale_trainable_op_from_config(
    op: "Ops.PlannedOperator", cfg
) -> "Ops.TrainableOperator":
    from sklearn.base import clone

    op_compat = op

    wrapped_op = clone(op_compat)
    cfg2 = smac_fixup_params(cfg)
    trainable = wrapped_op.with_params(**cfg2)
    return trainable


def get_smac_space(
    op: "Ops.PlannedOperator",
    lale_num_grids: Optional[float] = None,
    lale_pgo: Optional[PGO] = None,
    data_schema: Dict[str, Any] = {},
) -> ConfigurationSpace:
    """Top level function: given a lale operator, returns a ConfigurationSpace for use with SMAC.

    Parameters
    ----------
    op : The lale PlannedOperator
    lale_num_grids: integer or float, optional
        if set to an integer => 1, it will determine how many parameter grids will be returned (at most)
        if set to an float between 0 and 1, it will determine what fraction should be returned
        note that setting it to 1 is treated as in integer.  To return all results, use None
    """

    hp_grids = get_search_space_grids(
        op, num_grids=lale_num_grids, pgo=lale_pgo, data_schema=data_schema
    )
    cs = hp_grids_to_smac_cs(hp_grids)
    if should_print_search_space("true", "all", "backend", "smac"):
        name = op.name()
        if not name:
            name = "an operator"
        print(f"SMAC configuration for {name}:\n{str(cs)}")

    return cs


def smac_fixup_params(cfg):
    def strip_key(k: str) -> str:
        return k.rsplit("_", 1)[0]

    def transform_value(v):
        if v == "_lale_none":
            return None
        else:
            return v

    ret = {
        strip_key(k): transform_value(v)
        for (k, v) in cfg.get_dictionary().items()
        if k != "disjunct_discriminant"
    }
    return ret


# When sampling from distributions, this is the default number of samples to take.
# Users can override this by passing in num_samples to the appropriate function
SAMPLES_PER_DISTRIBUTION = 2

# We can first convert from our search space IR
# to a more limited grid structure
# This can than be converted to the format required for SMAC


def SearchSpaceNumberToSMAC(key: str, hp: SearchSpaceNumber) -> Hyperparameter:
    """Returns either a list of values intended to be sampled uniformly or a frozen scipy.stats distribution"""
    dist = "uniform"
    if hp.distribution:
        dist = hp.distribution
    if hp.maximum is None:
        raise ValueError(
            f"maximum not specified for a number with distribution {dist} for {key}"
        )
    max = hp.getInclusiveMax()
    if hp.minimum is None:
        raise ValueError(
            f"minimum not specified for a number with distribution {dist} for {key}"
        )
    min = hp.getInclusiveMin()

    log: bool
    if dist == "uniform" or dist == "integer":
        log = False
    elif dist == "loguniform":
        log = True
    else:
        raise ValueError(f"unknown/unsupported distribution {dist} for {key}")

    if hp.discrete:
        return UniformIntegerHyperparameter(key, min, max, log=log)
    else:
        return UniformFloatHyperparameter(key, min, max, log=log)


class FakeNone(object):
    pass


MyFakeNone = FakeNone()


def HPValuetoSMAC(key: str, hp: SearchSpace) -> Hyperparameter:
    def val_to_str(v):
        if v is None:
            return "_lale_none"
        else:
            return v

    if isinstance(hp, SearchSpaceEnum):
        return CategoricalHyperparameter(key, list(map(val_to_str, hp.vals)))
    elif isinstance(hp, SearchSpaceNumber):
        return SearchSpaceNumberToSMAC(key, hp)
    elif isinstance(hp, SearchSpaceArray):
        raise ValueError(
            f"Arrays are not yet supported by the SMAC backend (key: {key})"
        )
    else:
        raise ValueError(
            f"Not yet supported hp description ({type(hp)}) (key: {key}) in the GridSearchCV backend"
        )


def SearchSpaceGridtoSMAC(hp: SearchSpaceGrid, disc: int) -> Iterable[Hyperparameter]:
    return (HPValuetoSMAC(f"{k}_{disc}", v) for k, v in hp.items())


disc_str = "disjunct_discriminant"


def addSearchSpaceGrid(
    hp: SearchSpaceGrid, disc: int, parent_disc: Hyperparameter, cs: ConfigurationSpace
) -> None:
    smac = SearchSpaceGridtoSMAC(hp, disc)
    for hyp in smac:
        cs.add_hyperparameter(hyp)
        cs.add_condition(EqualsCondition(child=hyp, parent=parent_disc, value=disc))


def addSearchSpaceGrids(grids: List[SearchSpaceGrid], cs: ConfigurationSpace) -> None:
    parent_disc = CategoricalHyperparameter(disc_str, range(len(grids)))
    cs.add_hyperparameter(parent_disc)
    for (i, g) in enumerate(grids):
        addSearchSpaceGrid(g, i, parent_disc, cs)


def hp_grids_to_smac_cs(grids: List[SearchSpaceGrid]) -> ConfigurationSpace:
    cs: ConfigurationSpace = ConfigurationSpace()
    addSearchSpaceGrids(grids, cs)
    return cs



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import typing

import sklearn
import sklearn.pipeline

try:
    from sklearn.pipeline import if_delegate_has_method
except ImportError as e:
    import sklearn

    if sklearn.__version__ >= "1.0":
        from sklearn.utils.metaestimators import if_delegate_has_method
    else:
        raise e

import lale.docstrings
import lale.helpers
import lale.operators
from lale.schemas import Bool

logger = logging.getLogger(__name__)


class _PipelineImpl:
    def __init__(self, **hyperparams):
        if hyperparams.get("memory", None):
            logger.warning("Caching is not yet implemented.")
        if hyperparams.get("verbose", False):
            logger.warning(
                "Verbose is not implemented; instead, use lale.operators.logger.setLevel(logging.INFO)."
            )
        self._names = [name for name, _ in hyperparams["steps"]]
        new_steps = []
        for _, op in hyperparams["steps"]:
            if op is None or op == "passthrough":
                from lale.lib.lale import NoOp

                new_steps.append(NoOp)
            else:
                new_steps.append(op)
        self._pipeline = lale.operators.make_pipeline(*new_steps)
        self._final_estimator = self._pipeline.get_last()

    def fit(self, X, y=None, **fit_params):
        if y is None:
            self._pipeline = self._pipeline.fit(X, **fit_params)
        else:
            self._pipeline = self._pipeline.fit(X, y, **fit_params)
        self._final_estimator = self._pipeline.get_last()
        return self

    @if_delegate_has_method(delegate="_final_estimator")
    def predict(self, X, **predict_params):
        result = self._pipeline.predict(X, **predict_params)
        return result

    @if_delegate_has_method(delegate="_final_estimator")
    def predict_proba(self, X):
        result = self._pipeline.predict_proba(X)
        return result

    @if_delegate_has_method(delegate="_final_estimator")
    def transform(self, X, y=None):
        if y is None:
            result = self._pipeline.transform(X)
        else:
            result = self._pipeline.transform(X, y)
        return result

    def viz_label(self) -> str:
        return "Pipeline: " + ", ".join(self._names)


_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "additionalProperties": False,
            "required": ["steps"],
            "relevantToOptimizer": [],
            "properties": {
                "steps": {
                    "description": "List of (name, transform) tuples (implementing fit/transform) that are chained, in the order in which they are chained, with the last object an estimator.",
                    "type": "array",
                    "items": {
                        "description": "Tuple of (name, transform).",
                        "type": "array",
                        "laleType": "tuple",
                        "minItems": 2,
                        "maxItems": 2,
                        "items": [
                            {"description": "Name.", "type": "string"},
                            {
                                "anyOf": [
                                    {
                                        "description": "Transform.",
                                        "laleType": "operator",
                                    },
                                    {
                                        "description": "NoOp",
                                        "enum": [None, "passthrough"],
                                    },
                                ]
                            },
                        ],
                    },
                },
                "memory": {
                    "description": "Used to cache the fitted transformers of the pipeline.",
                    "anyOf": [
                        {
                            "description": "Path to the caching directory.",
                            "type": "string",
                        },
                        {
                            "description": "Object with the joblib.Memory interface",
                            "type": "object",
                            "forOptimizer": False,
                        },
                        {"description": "No caching.", "enum": [None]},
                    ],
                    "default": None,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {"description": "Features.", "laleType": "Any"},
        "y": {"description": "Target for supervised learning.", "laleType": "Any"},
    },
}

_input_predict_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {"X": {"description": "Features.", "laleType": "Any"}},
}

_output_predict_schema = {
    "description": "Predictions.",
    "laleType": "Any",
}

_input_predict_proba_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {"X": {"description": "Features.", "laleType": "Any"}},
}

_output_predict_proba_schema = {
    "description": "Probability of the sample for each class in the model.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {"X": {"description": "Features.", "laleType": "Any"}},
}

_output_transform_schema = {
    "description": "Features.",
    "laleType": "Any",
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Pipeline_ from scikit-learn creates a sequential list of operators.

.. _Pipeline: https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.pipeline.html",
    "import_from": "sklearn.pipeline",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


Pipeline = lale.operators.make_operator(_PipelineImpl, _combined_schemas)

if sklearn.__version__ >= "0.21":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.pipeline.Pipeline.html
    # new: https://scikit-learn.org/0.21/modules/generated/sklearn.pipeline.Pipeline.html
    Pipeline = typing.cast(
        lale.operators.PlannedIndividualOp,
        Pipeline.customize_schema(
            verbose=Bool(
                desc="If True, the time elapsed while fitting each step will be printed as it is completed.",
                default=False,
            ),
            set_as_available=True,
        ),
    )


lale.docstrings.set_docstrings(Pipeline)



# Copyright 2019, 2020, 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import unittest
from multiprocessing import cpu_count
from typing import cast

import numpy as np
import pandas as pd
from sklearn.datasets import make_regression
from sklearn.metrics import make_scorer
from sklearn.model_selection import TimeSeriesSplit

import lale.type_checking
from lale.datasets.uci import fetch_household_power_consumption
from lale.lib.autoai_ts_libs import (  # type: ignore # noqa; StandardRowMeanCenterMTS,; WindowTransformerMTS; DifferenceFlattenAutoEnsembler,; FlattenAutoEnsembler,; LocalizedFlattenAutoEnsembler,
    AutoaiTSPipeline,
    AutoaiWindowedWrappedRegressor,
    AutoaiWindowTransformedTargetRegressor,
    DifferenceFlattenAutoEnsembler,
    FlattenAutoEnsembler,
    LocalizedFlattenAutoEnsembler,
    MT2RForecaster,
    SmallDataWindowTargetTransformer,
    SmallDataWindowTransformer,
    StandardRowMeanCenter,
    T2RForecaster,
    WindowStandardRowMeanCenterMTS,
    WindowStandardRowMeanCenterUTS,
    cubic,
    flatten_iterative,
    linear,
)
from lale.lib.lale import Hyperopt
from lale.lib.sklearn import RandomForestRegressor, SimpleImputer


class TestAutoaiTSLibs(unittest.TestCase):
    def setUp(self):
        self.data = fetch_household_power_consumption()
        self.data = self.data.iloc[1::5000, :]
        self.X = self.data["Date"].to_numpy()
        self.y = self.data.drop(columns=["Date", "Time"]).applymap(
            lambda x: 0 if x == "?" else x
        )
        self.y = self.y.astype("float64").fillna(0).to_numpy()

    def doTestPipeline(
        self, trainable_pipeline, train_X, train_y, test_X, test_y, optimization=False
    ):
        def adjusted_smape(y_true, y_pred):
            """
            SMAPE
            """
            y_true, y_pred = np.array(y_true).ravel(), np.array(y_pred).ravel()
            if len(y_true) != len(y_pred):
                print(
                    "Size of Ground Truth and Predicted Values do not match!, returning None."
                )
                # May be raising error will interfere with daub execution if one pipeline fails
                # raise ValueError('Size of Ground Truth and Predicted Values do not match!')
                return None

            pred_diff = 2.0 * np.abs(cast(float, y_true - y_pred))
            divide = np.abs(y_true) + np.abs(y_pred)
            divide[divide < 1e-12] = 1.0
            scores = pred_diff / divide
            scores = np.array(scores, dtype=float)
            return np.nanmean(scores) * 100.0

        trained_pipeline = trainable_pipeline.fit(train_X, train_y)
        predicted = trained_pipeline.predict(test_X[:-1])
        if optimization:
            print(adjusted_smape(test_X[:-1], predicted))
        else:
            print(adjusted_smape(test_X[-1], predicted))
        with self.assertWarns(DeprecationWarning):
            trainable_pipeline.predict(train_X)
        trainable_pipeline.to_json()
        if optimization:
            hyperopt = Hyperopt(
                estimator=trainable_pipeline,
                max_evals=2,
                verbose=True,
                cv=TimeSeriesSplit(),
                scoring=make_scorer(adjusted_smape),
            )
            trained_hyperopt = hyperopt.fit(train_X, train_y)
            trained_hyperopt.predict(test_X)

    def test_pipeline_AWTTR_1(self):
        trainable = AutoaiTSPipeline(
            steps=[
                (
                    "AutoaiWindowTransformedTargetRegressor",
                    AutoaiWindowTransformedTargetRegressor(
                        regressor=SmallDataWindowTransformer()
                        >> SimpleImputer()
                        >> RandomForestRegressor()
                    ),
                )
            ]
        )
        self.doTestPipeline(trainable, self.y, self.y, self.y, self.y)

    def test_pipeline_AWTTR_2(self):
        trainable = AutoaiTSPipeline(
            steps=[
                (
                    "AutoaiWindowTransformedTargetRegressor",
                    AutoaiWindowTransformedTargetRegressor(
                        regressor=SmallDataWindowTransformer()
                        >> SimpleImputer()
                        >> RandomForestRegressor(),
                        estimator_prediction_type="rowwise",
                    ),
                )
            ]
        )
        self.doTestPipeline(
            trainable, self.y, self.y, self.y, self.y, optimization=True
        )

    def test_pipeline_SDWTT(self):
        trainable = AutoaiTSPipeline(
            steps=[
                (
                    "AutoaiWindowTransformedTargetRegressor",
                    AutoaiWindowTransformedTargetRegressor(
                        regressor=SmallDataWindowTargetTransformer(prediction_horizon=2)
                        >> SimpleImputer()
                        >> RandomForestRegressor(),
                        estimator_prediction_type="rowwise",
                    ),
                )
            ]
        )
        self.doTestPipeline(
            trainable, self.y, self.y, self.y, self.y, optimization=True
        )

    def test_pipeline_AWWR(self):
        trainable = AutoaiTSPipeline(
            steps=[
                (
                    "AutoaiWindowTransformedTargetRegressor",
                    AutoaiWindowedWrappedRegressor(
                        regressor=SmallDataWindowTransformer()
                        >> SimpleImputer()
                        >> RandomForestRegressor()
                    ),
                )
            ]
        )
        self.doTestPipeline(
            trainable, self.y, self.y, self.y, self.y, optimization=True
        )

    def test_pipeline_MT2RF(self):
        trainable = MT2RForecaster(target_columns=[0, 1, 2, 3, 4, 5, 6], trend="Linear")
        self.doTestPipeline(
            trainable, self.y, self.y, self.y, self.y, optimization=False
        )

    def test_pipeline_T2RF(self):
        trainable = T2RForecaster(trend="Linear")
        self.doTestPipeline(
            trainable,
            self.y[:, 0],
            self.y[:, 0],
            self.y[:, 0],
            self.y[:, 0],
            optimization=False,
        )


class StandardRowMeanCenterTest(unittest.TestCase):
    def setUp(self):
        infile = "https://vincentarelbundock.github.io/Rdatasets/csv/datasets/AirPassengers.csv"
        cols = ["ID", "time", "AirPassengers"]
        df = pd.read_csv(
            infile, names=cols, sep=r",", index_col="ID", engine="python", skiprows=1
        )
        trainnum = 100
        self.trainset = df.iloc[:trainnum, 1].values
        self.trainset = self.trainset.reshape(-1, 1)
        self.testset = df.iloc[trainnum:, 1].values
        self.testset = self.testset.reshape(-1, 1)

        self.lookback_window = 10
        self.prediction_horizon = 1

        # test corner case
        X, y = make_regression(
            n_features=10, n_informative=2, random_state=0, shuffle=False
        )
        self.X = X
        self.y = y

    def test_standard_row_mean_center_transformer(self):
        transformer = StandardRowMeanCenter()
        self.assertIsNotNone(transformer)
        trained = transformer.fit(self.trainset, self.trainset)
        (X_train, y_train) = trained.transform(self.trainset, self.trainset)
        self.assertTrue(X_train.shape[1] > 0)
        self.assertTrue(y_train.shape[1] > 0)

    def test_window_standard_row_mean_center_transformer_uts(self):
        transformer = WindowStandardRowMeanCenterUTS()
        self.assertIsNotNone(transformer)
        trained = transformer.fit(self.trainset, self.trainset)
        (X_train, y_train) = trained.transform(self.trainset, self.trainset)
        self.assertTrue(X_train.shape[1] > 0)
        self.assertTrue(y_train.shape[1] > 0)

    def test_window_standard_row_mean_center_transformer_mts(self):
        transformer = WindowStandardRowMeanCenterMTS()
        self.assertIsNotNone(transformer)

        trained = transformer.fit(self.X, self.X)
        (X_train, y_train) = trained.transform(self.trainset, self.trainset)
        self.assertTrue(X_train.shape[1] > 0)
        self.assertTrue(y_train.shape[1] > 0)

    def tearDown(self):
        pass


class TimeseriesWindowTransformerTest(unittest.TestCase):
    def setUp(self):
        infile = "https://vincentarelbundock.github.io/Rdatasets/csv/datasets/AirPassengers.csv"
        cols = ["ID", "time", "AirPassengers"]
        df = pd.read_csv(
            infile, names=cols, sep=r",", index_col="ID", engine="python", skiprows=1
        )
        trainnum = 100
        self.trainset = df.iloc[:trainnum, 1].values
        self.trainset = self.trainset.reshape(-1, 1)
        self.testset = df.iloc[trainnum:, 1].values
        self.testset = self.testset.reshape(-1, 1)

        self.lookback_window = 10
        self.prediction_horizon = 1

        # test corner case
        X, y = make_regression(
            n_features=10, n_informative=2, random_state=0, shuffle=False
        )
        self.X = X
        self.y = y

    def check_transformer(self, transformer, X):
        tr = transformer.fit(X)
        self.assertIsNotNone(tr)
        Xt = tr.transform(X)
        self.assertEqual(X.shape[0], Xt.shape[0])
        return Xt

    def test_small_data_window_transformer(self):
        transformer = SmallDataWindowTransformer(lookback_window=None)
        self.assertIsNotNone(transformer)
        Xt = self.check_transformer(transformer=transformer, X=self.trainset)
        self.assertTrue(Xt.shape[1] > 0)

        transformer = SmallDataWindowTransformer(
            lookback_window=self.lookback_window, cache_last_window_trainset=True
        )
        self.assertIsNotNone(transformer)
        Xt = self.check_transformer(transformer=transformer, X=self.trainset)
        self.assertTrue(Xt.shape[1] > 0)

        Xtest = transformer.transform(X=self.testset)
        self.assertTrue(Xtest.shape[1] > 0)

        transformer = SmallDataWindowTransformer(lookback_window=200)
        Xt = self.check_transformer(transformer=transformer, X=self.trainset)
        self.assertTrue(Xt.shape[1] > 0)

        transformer = SmallDataWindowTransformer(lookback_window=None)
        ftransformer = transformer.fit(X=self.X)
        self.assertIsNotNone(ftransformer)

    def test_small_data_window_target_transformer(self):
        transformer = SmallDataWindowTargetTransformer(
            prediction_horizon=self.prediction_horizon
        )
        self.assertIsNotNone(transformer)
        Yt = self.check_transformer(transformer=transformer, X=self.trainset)
        self.assertEqual(np.count_nonzero(np.isnan(Yt)), 1)

    # def test_tested(self):
    #     self.assertTrue(False, "this module was tested")

    def tearDown(self):
        pass


class TestMT2RForecaster(unittest.TestCase):
    """class for testing MT2RForecaster"""

    @classmethod
    def setUp(cls):
        x = np.arange(30)
        y = np.arange(300, 330)
        X = np.array([x, y])
        X = np.transpose(X)
        cls.target_columns = [0, 1]
        cls.X = X

    def test_fit(self):
        """method for testing the fit method of MT2RForecaster"""
        test_class = self.__class__
        model = MT2RForecaster(target_columns=test_class.target_columns)
        model.fit(test_class.X)

    def test_predict_multicols(self):
        """Tests the multivariate predict method of MT2RForecaster"""
        test_class = self.__class__
        model = MT2RForecaster(
            target_columns=test_class.target_columns, prediction_win=2
        )
        fitted_model = model.fit(test_class.X)
        ypred = fitted_model.predict()
        self.assertEqual(len(ypred), 2)
        model = MT2RForecaster(
            target_columns=test_class.target_columns, prediction_win=1
        )
        fitted_model = model.fit(test_class.X)
        ypred = fitted_model.predict()
        self.assertEqual(len(ypred), 1)
        model = MT2RForecaster(target_columns=test_class.target_columns, trend="Mean")
        fitted_model = model.fit(test_class.X)
        ypred = fitted_model.predict()
        self.assertEqual(len(ypred), 12)
        model = MT2RForecaster(target_columns=test_class.target_columns, trend="Poly")
        fitted_model = model.fit(test_class.X)
        ypred = fitted_model.predict()
        self.assertEqual(len(ypred), 12)

    def test_predict_prob(self):
        """Tests predict_prob method of MT2RForecaster"""
        test_class = self.__class__
        model = MT2RForecaster(target_columns=[0], prediction_win=2)
        fitted_model = model.fit(test_class.X)
        ypred = fitted_model.predict_proba()
        assert ypred is not None
        self.assertEqual(ypred.shape, (2, 1, 2))
        model = MT2RForecaster(target_columns=[0, 1], prediction_win=2)
        fitted_model = model.fit(test_class.X)
        ypred = fitted_model.predict_proba()
        self.assertIsNotNone(ypred)
        assert ypred is not None
        self.assertEqual(ypred.shape, (2, 2, 2))

    def test_predict_uni_cols(self):
        """Tests the univariate predict method of MT2RForecaster"""
        x = np.arange(10)
        X = x.reshape(-1, 1)
        model = MT2RForecaster(target_columns=[0], prediction_win=2)
        fitted_model = model.fit(X)
        ypred = fitted_model.predict()
        self.assertEqual(len(ypred), 2)
        model = MT2RForecaster(target_columns=[0], prediction_win=1)
        fitted_model = model.fit(X)
        ypred = fitted_model.predict()
        self.assertEqual(len(ypred), 1)


class TestPrettyPrint(unittest.TestCase):
    def test_param_grid(self):
        printed1 = """from autoai_ts_libs.srom.estimators.time_series.models.srom_estimators import (
    LocalizedFlattenAutoEnsembler,
)
from autoai_ts_libs.srom.estimators.regression.auto_ensemble_regressor import (
    EnsembleRegressor,
)
from autoai_ts_libs.srom.joint_optimizers.auto.auto_regression import (
    AutoRegression,
)
from sklearn.multioutput import MultiOutputRegressor
from sklearn.linear_model import LinearRegression
import autoai_ts_libs.srom.joint_optimizers.cv.time_series_splits
import autoai_ts_libs.srom.joint_optimizers.pipeline.srom_param_grid
import sklearn.metrics
import autoai_ts_libs.srom.joint_optimizers.utils.no_op
import sklearn.preprocessing
import sklearn.multioutput
import sklearn.linear_model
import xgboost.sklearn
import lale

lale.wrap_imported_operators()
linear_regression = LinearRegression(n_jobs=1)
multi_output_regressor = MultiOutputRegressor(
    estimator=linear_regression, n_jobs=5
)
auto_regression = AutoRegression(
    cv=autoai_ts_libs.srom.joint_optimizers.cv.time_series_splits.TimeSeriesTrainTestSplit(
        n_splits=1, n_test_size=423, overlap_len=0
    ),
    execution_time_per_pipeline=3,
    level="default",
    param_grid=autoai_ts_libs.srom.joint_optimizers.pipeline.srom_param_grid.SROMParamGrid(),
    scoring=sklearn.metrics.make_scorer(
        sklearn.metrics.mean_absolute_error, greater_is_better=False
    ),
    stages=[
        [
            (
                "skipscaling",
                autoai_ts_libs.srom.joint_optimizers.utils.no_op.NoOp(),
            ),
            ("minmaxscaler", sklearn.preprocessing.MinMaxScaler()),
        ],
        [
            (
                "molinearregression",
                sklearn.multioutput.MultiOutputRegressor(
                    estimator=sklearn.linear_model.LinearRegression(n_jobs=1),
                    n_jobs=5,
                ),
            ),
            (
                "mosgdregressor",
                sklearn.multioutput.MultiOutputRegressor(
                    estimator=sklearn.linear_model.SGDRegressor(
                        random_state=0
                    ),
                    n_jobs=5,
                ),
            ),
            (
                "moxgbregressor",
                sklearn.multioutput.MultiOutputRegressor(
                    estimator=xgboost.sklearn.XGBRegressor(
                        missing=float("nan"), objective="reg:squarederror"
                    ),
                    n_jobs=5,
                ),
            ),
        ],
    ],
    total_execution_time=3,
    best_estimator_so_far=multi_output_regressor,
)
ensemble_regressor = EnsembleRegressor(
    cv=None,
    execution_platform=None,
    execution_time_per_pipeline=None,
    level=None,
    n_estimators_for_pred_interval=1,
    n_leaders_for_ensemble=1,
    num_option_per_pipeline_for_intelligent_search=None,
    num_options_per_pipeline_for_random_search=None,
    save_prefix=None,
    total_execution_time=None,
    auto_regression=auto_regression,
)
pipeline = LocalizedFlattenAutoEnsembler(
    feature_columns=[0, 1],
    target_columns=[0, 1],
    lookback_win=10,
    pred_win=5,
    dag_granularity="multioutput_flat",
    execution_time_per_pipeline=3,
    init_time_optimization=True,
    multistep_prediction_strategy="multioutput",
    multistep_prediction_win=5,
    n_estimators_for_pred_interval=1,
    n_jobs=5,
    n_leaders_for_ensemble=1,
    store_lookback_history=True,
    total_execution_time=3,
    estimator=ensemble_regressor,
)"""
        globals2 = {}
        locals2 = {}
        exec(printed1, globals2, locals2)
        pipeline2 = locals2["pipeline"]
        printed2 = pipeline2.pretty_print()
        self.maxDiff = None
        self.assertEqual(printed1, printed2)


class TestWatForeForecasters(unittest.TestCase):
    def setUp(self):
        print(
            "..................................Watfore TSLIB tests.................................."
        )

    @unittest.skip("Does not work, not sure if it is supposed to work.")
    def test_watfore_pickle_write(self):
        print(
            "................................Watfore TSLIB Training and Pickle Write.................................."
        )
        # for this make sure sys.modules['ai4ml_ts.estimators'] = watfore is removed from init otherwise package is confused
        import pickle

        from lale.lib.autoai_ts_libs import WatForeForecaster

        fr = WatForeForecaster(
            algorithm="hw", samples_per_season=0.6
        )  # , initial_training_seasons=3
        ts_val = [
            [0, 10.0],
            [1, 20.0],
            [2, 30.0],
            [3, 40.0],
            [4, 50.0],
            [5, 60.0],
            [6, 70.0],
            [7, 31.0],
            [8, 80.0],
            [9, 90.0],
        ]

        ml = fr.fit(ts_val, None)
        pr_before = fr.predict(None)
        fr2 = None
        print("Predictions before saving", pr_before)
        f_name = "./lale/datasets/autoai/watfore_pipeline.pickle"  # ./tests/watfore/watfore_pipeline.pickle
        with open(f_name, "wb") as pkl_dump:
            pickle.dump(ml, pkl_dump)
        pr2 = fr.predict(None)
        print("Predictions after pikcle dump", pr2)
        self.assertTrue(
            (np.asarray(pr2).ravel() == np.asarray(pr_before).ravel()).all()
        )
        print("Pickle Done. Now loading...")
        with open(f_name, "rb") as pkl_dump:
            fr2 = pickle.load(pkl_dump)
        if fr2 is not None:
            preds = fr2.predict(None)
            print("Predictions after loading", preds)
            self.assertTrue(len(preds) == 1)
            self.assertTrue(
                (np.asarray(preds).ravel() == np.asarray(pr_before).ravel()).all()
            )
        else:
            print("Failed to Load model(s) from location" + f_name)
            self.fail("Failed to Load model(s) from location" + f_name)
        print(
            "................................Watfor TSLIB Pickle Write Done........................."
        )

    @unittest.skip("Does not work, not sure if it is supposed to work.")
    def test_load_watfore_pipeline(self):

        print(
            "..................................Watfore TSLIB Pickle load and Predict................................."
        )
        import pickle

        # f_name = './tests/watfore/watfore_pipeline.pickle'
        f_name = "watfore_pipeline.pickle"
        print("Pickle Done. Now loading...")
        with open(f_name, "rb") as pkl_dump:
            fr2 = pickle.load(pkl_dump)

        if fr2 is not None:
            preds = fr2.predict(None)
            print("Predictions after loading", preds)
            self.assertTrue(len(preds) == 1)

            # self.assertTrue((np.asarray(preds).ravel() == np.asarray(pr_before).ravel()).all())
        else:
            print("Failed to Load model(s) from location" + f_name)
            self.fail("Failed to Load model(s) from location" + f_name)
        print(
            "................................Watfore TSLIB Read and predict Done........................."
        )


class TestImportExport(unittest.TestCase):
    def setUp(self):
        self.data = fetch_household_power_consumption()
        self.data = self.data.iloc[1::5000, :]
        self.X = self.data["Date"].to_numpy()
        self.y = self.data.drop(columns=["Date", "Time"]).applymap(
            lambda x: 0 if x == "?" else x
        )
        self.y = self.y.astype("float64").fillna(0).to_numpy()


def train_test_split(inputX, split_size):
    return inputX[:split_size], inputX[split_size:]


def get_srom_time_series_estimators(
    feature_column_indices,
    target_column_indices,
    lookback_window,
    prediction_horizon,
    time_column_index=-1,
    optimization_stetagy="Once",
    mode="Test",
    number_rows=None,
    n_jobs=cpu_count() - 1,
):
    """
    This method returns the best performing time_series estimators in SROM. The no. of estiamtors
    depend upon the mode of operation. The mode available are 'Test', 'Benchmark' and
    'benchmark_extended'.
    Parameters:
        feature_column_indices (list): feature indices.
        target_column_indices (list): target indices.
        time_column_index (int): time column index.
        lookback_window (int): Look-back window for the models returned.
        prediction_horizon (int): Look-ahead window for the models returned.
        init_time_optimization (string , optional): whether to optimize at the start of automation.
        mode (string, optional) : The available modes are test, benchmark, benchmark_extended.
    """
    srom_estimators = []
    from lale.lib.autoai_ts_libs import EnsembleRegressor

    # adding P21
    srom_estimators.append(
        MT2RForecaster(
            target_columns=target_column_indices,
            trend="Linear",
            lookback_win=lookback_window,
            prediction_win=prediction_horizon,
            n_jobs=n_jobs,
        )
    )

    ensemble_regressor = EnsembleRegressor(
        cv=None,
        execution_platform=None,
        execution_time_per_pipeline=None,
        level=None,
        n_estimators_for_pred_interval=1,
        n_leaders_for_ensemble=1,
        num_option_per_pipeline_for_intelligent_search=None,
        num_options_per_pipeline_for_random_search=None,
        save_prefix=None,
        total_execution_time=None,
    )
    # Setting commong parameters
    auto_est_params = {
        "feature_columns": feature_column_indices,
        "target_columns": target_column_indices,
        "lookback_win": lookback_window,
        "pred_win": prediction_horizon,
        "time_column": time_column_index,
        "execution_platform": "spark_node_random_search",
        "n_leaders_for_ensemble": 1,
        "n_estimators_for_pred_interval": 1,
        "max_samples_for_pred_interval": 1.0,
        "init_time_optimization": True,
        "dag_granularity": "flat",
        "total_execution_time": 3,
        "execution_time_per_pipeline": 3,
        "store_lookback_history": True,
        "n_jobs": n_jobs,
        "estimator": ensemble_regressor,
    }

    if prediction_horizon > 1:
        auto_est_params["multistep_prediction_win"] = prediction_horizon
        auto_est_params["multistep_prediction_strategy"] = "multioutput"
        auto_est_params["dag_granularity"] = "multioutput_flat"
    elif len(target_column_indices) > 1:
        auto_est_params["dag_granularity"] = "multioutput_flat"
    else:
        pass
    auto_est_params["data_transformation_scheme"] = "log"

    # adding P18
    P18_params = copy.deepcopy(auto_est_params)
    srom_estimators.append(FlattenAutoEnsembler(**P18_params))

    # adding P17, Local Model
    P17_params = copy.deepcopy(auto_est_params)
    srom_estimators.append(DifferenceFlattenAutoEnsembler(**P17_params))

    # adding P14m Local Model
    auto_est_params["data_transformation_scheme"] = None
    P14_params = copy.deepcopy(auto_est_params)
    srom_estimators.append(LocalizedFlattenAutoEnsembler(**P14_params))

    return srom_estimators


class TestSROMEnsemblers(unittest.TestCase):
    """Test various SROM Ensemblers classes"""

    @unittest.skip(
        "Does not work as the fit complains that there is no best_estimator_so_far"
    )
    def test_fit_predict_predict_sliding_window_univariate_single_step(self):
        X = np.arange(1, 441)
        X = X.reshape(-1, 1)
        SIZE = len(X)
        target_columns = [0]
        number_rows = SIZE
        prediction_horizon = 1
        lookback_window = 10
        run_mode = "test"

        srom_estimators = get_srom_time_series_estimators(
            feature_column_indices=target_columns,
            target_column_indices=target_columns,
            lookback_window=lookback_window,
            prediction_horizon=prediction_horizon,
            optimization_stetagy="Once",
            mode=run_mode,
            number_rows=number_rows,
        )
        for index, estimator in enumerate(srom_estimators[1:]):
            X_train, X_test = train_test_split(
                X, SIZE - (prediction_horizon + lookback_window)
            )
            estimator.fit(X_train)
            y_pred = estimator.predict(X_test)
            assert len(y_pred) == prediction_horizon
            assert y_pred.shape[1] == len(target_columns)
            y_pred_win = estimator.predict_sliding_window(X_test)
            assert len(y_pred_win) == lookback_window + 1
            assert y_pred_win.shape[1] == len(target_columns)

    @unittest.skip(
        "Does not work as the fit complains that there is no best_estimator_so_far"
    )
    def test_fit_predict_predict_sliding_window_univariate_multi_step(self):
        X = np.arange(1, 441)
        X = X.reshape(-1, 1)
        SIZE = len(X)
        target_columns = [0]
        number_rows = SIZE
        prediction_horizon = 8
        lookback_window = 10
        run_mode = "test"

        srom_estimators = get_srom_time_series_estimators(
            feature_column_indices=target_columns,
            target_column_indices=target_columns,
            lookback_window=lookback_window,
            prediction_horizon=prediction_horizon,
            optimization_stetagy="Once",
            mode=run_mode,
            number_rows=number_rows,
        )
        for index, estimator in enumerate(srom_estimators[1:]):
            X_train, X_test = train_test_split(
                X, SIZE - (prediction_horizon + lookback_window)
            )
            estimator.fit(X_train)
            y_pred = estimator.predict(X_test)
            assert len(y_pred) == prediction_horizon
            assert y_pred.shape[1] == len(target_columns)
            y_pred_win = estimator.predict_multi_step_sliding_window(X_test)
            assert y_pred_win.shape[1] == len(target_columns)

    @unittest.skip(
        "Does not work as the fit complains that there is no best_estimator_so_far"
    )
    def test_fit_predict_predict_sliding_window_multivariate_single_step(self):
        X = np.arange(1, 441)
        X = X.reshape(-1, 1)
        X2 = np.arange(1001, 1441)
        X2 = X2.reshape(-1, 1)
        X3 = np.arange(10001, 10441)
        X3 = X3.reshape(-1, 1)
        X = np.hstack([X, X2, X3])
        SIZE = len(X)
        target_columns = [0, 1, 2]
        number_rows = SIZE
        prediction_horizon = 1
        lookback_window = 10
        run_mode = "test"
        srom_estimators = get_srom_time_series_estimators(
            feature_column_indices=target_columns,
            target_column_indices=target_columns,
            lookback_window=lookback_window,
            prediction_horizon=prediction_horizon,
            optimization_stetagy="Once",
            mode=run_mode,
            number_rows=number_rows,
        )
        for index, estimator in enumerate(srom_estimators[1:]):
            X_train, X_test = train_test_split(
                X, SIZE - (prediction_horizon + lookback_window)
            )
            estimator.fit(X_train)
            y_pred = estimator.predict(X_test)
            assert len(y_pred) == prediction_horizon
            assert y_pred.shape[1] == len(target_columns)
            y_pred_win = estimator.predict_sliding_window(X_test)
            assert len(y_pred_win) == lookback_window + 1
            assert y_pred_win.shape[1] == len(target_columns)

    @unittest.skip(
        "Does not work as the fit complains that there is no best_estimator_so_far"
    )
    def test_fit_predict_predict_sliding_window_multivariate_multi_step(self):
        X = np.arange(1, 441)
        X = X.reshape(-1, 1)
        X2 = np.arange(1001, 1441)
        X2 = X2.reshape(-1, 1)
        X3 = np.arange(10001, 10441)
        X3 = X3.reshape(-1, 1)
        X = np.hstack([X, X2, X3])
        SIZE = len(X)
        target_columns = [0, 1, 2]
        number_rows = SIZE
        prediction_horizon = 8
        lookback_window = 10
        run_mode = "test"
        srom_estimators = get_srom_time_series_estimators(
            feature_column_indices=target_columns,
            target_column_indices=target_columns,
            lookback_window=lookback_window,
            prediction_horizon=prediction_horizon,
            optimization_stetagy="Once",
            mode=run_mode,
            number_rows=number_rows,
        )
        for index, estimator in enumerate(srom_estimators[1:]):
            X_train, X_test = train_test_split(
                X, SIZE - (prediction_horizon + lookback_window)
            )
            estimator.fit(X_train)
            y_pred = estimator.predict(X_test)
            assert len(y_pred) == prediction_horizon
            assert y_pred.shape[1] == len(target_columns)
            y_pred_win = estimator.predict_multi_step_sliding_window(X_test)
            assert y_pred_win.shape[1] == len(target_columns)


class TestInterpolatorImputers(unittest.TestCase):
    """class for testing different time-series imputers."""

    @classmethod
    def setUp(cls):
        uni_x = pd.DataFrame({"A": [12, 4, 5, None, 1]})
        multi_x = pd.DataFrame(
            {
                "A": [12, 4, 5, None, 1],
                "B": [None, 2, 54, 3, None],
                "C": [20, 16, None, 3, 8],
                "D": [14, 3, None, None, 6],
            }
        )
        cls.uni_x = uni_x
        cls.multi_x = multi_x

    def test_linear_imputer(self):
        """Test  LinearImputer"""
        test_class = self.__class__
        uni_x = test_class.uni_x
        multi_x = test_class.multi_x

        # Test univariate
        imputer = linear().convert_to_trained()
        X_tf = imputer.transform(uni_x)
        self.assertAlmostEqual(np.mean(X_tf), 5.0, 2)

        # Test multivariate
        imputer = linear().convert_to_trained()
        X_tf = imputer.transform(multi_x)
        self.assertAlmostEqual(np.mean(X_tf), 8.875, 2)

        # Test numpy
        imputer = linear().convert_to_trained()
        X_tf = imputer.transform(uni_x.values)
        self.assertAlmostEqual(np.mean(X_tf), 5.0, 2)

        # Test numpy
        imputer = linear().convert_to_trained()
        X_tf = imputer.transform(multi_x.values)
        self.assertAlmostEqual(np.mean(X_tf), 8.875, 2)

    def test_cubic_imputer(self):
        """Test  CubicImputer"""
        test_class = self.__class__
        uni_x = test_class.uni_x
        multi_x = test_class.multi_x[["A", "C"]]

        # Test univariate
        imputer = cubic().convert_to_trained()
        X_tf = imputer.transform(uni_x)
        self.assertAlmostEqual(np.mean(X_tf), 5.75, 2)

        # Test multivariate
        imputer = cubic().convert_to_trained()
        X_tf = imputer.transform(multi_x)
        self.assertAlmostEqual(np.mean(X_tf), 8.375, 2)

        # Test numpy
        imputer = cubic().convert_to_trained()
        X_tf = imputer.transform(uni_x.values)
        self.assertAlmostEqual(np.mean(X_tf), 5.75, 2)

        # Test numpy
        imputer = cubic().convert_to_trained()
        X_tf = imputer.transform(multi_x.values)
        self.assertAlmostEqual(np.mean(X_tf), 8.375, 2)


class TestFlattenImputers(unittest.TestCase):
    """class for testing different Flatten imputers"""

    @classmethod
    def setUp(cls):
        uni_x = pd.DataFrame({"A": [1, 2, 3, 4, 5, None, 7, 8, 9]})
        multi_x = pd.DataFrame(
            {
                "A": [1, 2, 3, 4, 5, None, 7, 8, 9, 10],
                "B": [101, 102, None, 104, 105, 106, 107, 108, 109, 110],
                "C": [51, 52, None, 54, 55, 56, None, 58, 59, 60],
            }
        )
        imputers = [flatten_iterative]
        cls.uni_x = uni_x
        cls.multi_x = multi_x
        cls.imputers = imputers

    def test_fit_transform_flatten_imputers(self):
        """
        Test Fit and transform flatten imputers
        """
        test_class = self.__class__
        uni_x = test_class.uni_x
        multi_x = test_class.multi_x

        # Test univariate timeseries data
        est = flatten_iterative()
        try:
            for est in test_class.imputers:
                print("testing", est)
                imputer = est(order=5)
                imputer.fit(uni_x)
                interpolated = imputer.transform(uni_x)
                self.assertFalse(np.any(np.isnan(interpolated)))
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))

        # Test multivariate timeseries data
        try:
            for est in test_class.imputers:
                print("testing", est)
                imputer = est(order=5)
                imputer.fit(multi_x)
                interpolated = imputer.transform(multi_x)
                self.assertFalse(np.any(np.isnan(interpolated)))
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))

    def test_fit_transform_flatten_imputers_without_nan(self):
        """
        Test Fit and transform flatten imputers with nan
        """
        test_class = self.__class__
        uni_x = pd.DataFrame({"A": [1, 2, 3, 4, 5, 6, 7, 8, 9]})
        multi_x = pd.DataFrame(
            {
                "A": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                "B": [101, 102, 103, 104, 105, 106, 107, 108, 109, 110],
                "C": [51, 52, 53, 54, 55, 56, 57, 58, 59, 60],
            }
        )
        est = flatten_iterative()
        # Test univariate timeseries data
        try:
            for est in test_class.imputers:
                print("testing", est)
                imputer = est(order=5)
                imputer.fit(uni_x)
                interpolated = imputer.transform(uni_x)
                self.assertFalse(np.any(np.isnan(interpolated)))
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))

        # Test multivariate timeseries data
        try:
            for est in test_class.imputers:
                print("testing", est)
                imputer = est(order=5)
                imputer.fit(multi_x)
                interpolated = imputer.transform(multi_x)
                self.assertFalse(np.any(np.isnan(interpolated)))
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))

    @unittest.skip("Need to fix.")
    def test_set_params(self):
        """
        Test set_params
        """
        test_class = self.__class__
        PARAMS = {
            flatten_iterative: {
                "base_imputer__random_state": 24,
            }
        }
        est = flatten_iterative()
        try:
            for est in test_class.imputers[0:1]:
                params = PARAMS[est]
                imputer = est()
                imputer.set_params(**params)
                self.assertEqual(
                    PARAMS[est]["base_imputer__random_state"],
                    imputer.base_imputer.get_params()["random_state"],
                )
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))

    def test_get_params(self):
        """
        Test get_params
        """
        test_class = self.__class__
        est = flatten_iterative()
        try:
            for est in test_class.imputers[0:1]:
                imputer = est()
                self.assertIsNotNone(imputer.get_params())
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))

    def test_fit_transform_flatten_imputers_with_less_data(self):
        """
        Test Fit and transform flatten imputers with less data
        """
        test_class = self.__class__
        est = flatten_iterative()
        multi_x = pd.DataFrame(
            {
                "A": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                "B": [101, 102, 103, 104, 105, 106, 107, 108, 109, 110],
                "C": [51, 52, 53, 54, 55, 56, 57, 58, 59, 60],
            }
        )
        test_x = pd.DataFrame(
            {
                "A": [7, 8, 9, 10],
                "B": [107, 108, 109, 110],
                "C": [57, 58, 59, 60],
            }
        )
        # Test multivariate timeseries data
        try:
            for est in test_class.imputers:
                print("testing", est)
                imputer = est(order=5)
                imputer.fit(multi_x)
                interpolated = imputer.transform(test_x)
                self.assertFalse(np.any(np.isnan(interpolated)))
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))
        test_x = pd.DataFrame(
            {
                "A": [7, None, 9, 10],
                "B": [107, None, 109, 110],
                "C": [57, 58, 59, 60],
            }
        )
        try:
            for est in test_class.imputers:
                print("testing", est)
                imputer = est(order=5)
                imputer.fit(multi_x)
                self.assertRaises(Exception, imputer, "transform", test_x)
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))

    def test_fit_transform_flatten_imputers_with_other_missing_value(self):
        """
        Test Fit and transform flatten imputers
        """
        test_class = self.__class__
        uni_x = pd.DataFrame({"A": [1, 2, 3, 4, 5, -999, 7, 8, 9]})
        multi_x = pd.DataFrame(
            {
                "A": [1, 2, 3, 4, 5, -999, 7, 8, 9, 10],
                "B": [101, 102, -999, 104, 105, 106, 107, 108, 109, 110],
                "C": [51, 52, -999, 54, 55, 56, -999, 58, 59, 60],
            }
        )
        est = flatten_iterative()
        # Test univariate timeseries data
        try:
            for est in test_class.imputers:
                print("testing", est)
                imputer = est(order=5, missing_val_identifier=-999)
                imputer.fit(uni_x)
                interpolated = imputer.transform(uni_x)
                self.assertFalse(np.any(np.isnan(interpolated)))
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))

        # Test multivariate timeseries data
        try:
            for est in test_class.imputers:
                print("testing", est)
                imputer = est(order=5, missing_val_identifier=-999)
                imputer.fit(multi_x)
                interpolated = imputer.transform(multi_x)
                self.assertFalse(np.any(np.isnan(interpolated)))
        except Exception as e:
            self.fail("Failed : " + str(est.name()) + " " + str(e))


class TestSchemas(unittest.TestCase):
    def setUp(self):
        pass


def create_function_test_schemas(obj_name):
    def test_schemas(self):
        import importlib

        module_name = ".".join(obj_name.split(".")[0:-1])
        class_name = obj_name.split(".")[-1]
        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        if class_name == "MT2RForecaster":
            obj = class_(target_columns=[0])
        else:
            obj = class_()

        obj._check_schemas()
        # test_schemas_are_schemas
        lale.type_checking.validate_is_schema(obj.hyperparam_schema())

    test_schemas.__name__ = "test_{0}".format(obj.split(".")[-1])
    return test_schemas


objs = [
    # "lale.lib.autoai_ts_libs.AutoaiTSPipeline",  # does not work as steps can be None or empty
    "lale.lib.autoai_ts_libs.AutoaiWindowedWrappedRegressor",
    "lale.lib.autoai_ts_libs.AutoaiWindowTransformedTargetRegressor",
    "lale.lib.autoai_ts_libs.MT2RForecaster",
    "lale.lib.autoai_ts_libs.SmallDataWindowTargetTransformer",
    "lale.lib.autoai_ts_libs.SmallDataWindowTransformer",
    "lale.lib.autoai_ts_libs.StandardRowMeanCenter",
    "lale.lib.autoai_ts_libs.T2RForecaster",
    "lale.lib.autoai_ts_libs.WindowStandardRowMeanCenterMTS",
    "lale.lib.autoai_ts_libs.cubic",
    "lale.lib.autoai_ts_libs.flatten_iterative",
    "lale.lib.autoai_ts_libs.linear",
    "lale.lib.autoai_ts_libs.fill",
    "lale.lib.autoai_ts_libs.previous",
    "lale.lib.autoai_ts_libs.next",
    "lale.lib.autoai_ts_libs.AutoRegression",
    "lale.lib.autoai_ts_libs.EnsembleRegressor",
    "lale.lib.autoai_ts_libs.WatForeForecaster",
]
for obj in objs:
    setattr(
        TestSchemas,
        "test_{0}".format(obj.split(".")[-1]),
        create_function_test_schemas(obj),
    )



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Tuple, Type

import numpy as np
import pandas as pd
import scipy.sparse

import lale.type_checking
from lale.helpers import _is_spark_df
from lale.type_checking import JSON_TYPE

try:
    import torch

    torch_installed = True
except ImportError:
    torch_installed = False

try:
    import py4j.protocol
    import pyspark.sql

    spark_installed = True
except ImportError:
    spark_installed = False


# See instructions for subclassing numpy ndarray:
# https://docs.scipy.org/doc/numpy/user/basics.subclassing.html
class NDArrayWithSchema(np.ndarray):
    def __new__(
        cls,
        shape,
        dtype=float,
        buffer=None,
        offset=0,
        strides=None,
        order=None,
        json_schema=None,
        table_name=None,
    ):
        result = super(NDArrayWithSchema, cls).__new__(
            cls, shape, dtype, buffer, offset, strides, order  # type: ignore
        )
        result.json_schema = json_schema
        result.table_name = table_name
        return result

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.json_schema = getattr(obj, "json_schema", None)
        self.table_name = getattr(obj, "table_name", None)


# See instructions for subclassing pandas DataFrame:
# https://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas
class DataFrameWithSchema(pd.DataFrame):
    _internal_names = pd.DataFrame._internal_names + ["json_schema", "table_name"]
    _internal_names_set = set(_internal_names)

    @property
    def _constructor(self):
        return DataFrameWithSchema


class SeriesWithSchema(pd.Series):
    _internal_names = pd.DataFrame._internal_names + [
        "json_schema",
        "table_name",
        "folds_for_monoid",
    ]
    _internal_names_set = set(_internal_names)

    @property
    def _constructor(self):
        return SeriesWithSchema


def add_schema(obj, schema=None, raise_on_failure=False, recalc=False) -> Any:
    from lale.settings import disable_data_schema_validation

    if disable_data_schema_validation:
        return obj
    if obj is None:
        return None
    if isinstance(obj, NDArrayWithSchema):
        result = obj
    elif isinstance(obj, np.ndarray):
        result = obj.view(NDArrayWithSchema)
    elif isinstance(obj, SeriesWithSchema):
        result = obj
    elif isinstance(obj, pd.Series):
        result = SeriesWithSchema(obj)
    elif isinstance(obj, DataFrameWithSchema):
        result = obj
    elif isinstance(obj, pd.DataFrame):
        result = DataFrameWithSchema(obj)
    elif is_list_tensor(obj):
        obj = np.array(obj)
        result = obj.view(NDArrayWithSchema)
    elif raise_on_failure:
        raise ValueError(f"unexpected type(obj) {type(obj)}")
    else:
        return obj
    if recalc:
        setattr(result, "json_schema", None)
    if getattr(result, "json_schema", None) is None:
        if schema is None:
            setattr(result, "json_schema", to_schema(obj))
        else:
            lale.type_checking.validate_is_schema(schema)
            setattr(result, "json_schema", schema)
    return result


def add_schema_adjusting_n_rows(obj, schema):
    assert isinstance(obj, (np.ndarray, pd.DataFrame, pd.Series)), type(obj)
    assert schema.get("type", None) == "array", schema
    n_rows = obj.shape[0]
    mod_schema = {**schema, "minItems": n_rows, "maxItems": n_rows}
    result = add_schema(obj, mod_schema)
    return result


def add_table_name(obj, name) -> Any:
    if obj is None:
        return None
    if name is None:
        return obj
    if spark_installed and isinstance(obj, pyspark.sql.DataFrame):
        # alias method documentation: https://spark.apache.org/docs/latest/api/python/reference/api/pyspark.sql.DataFrame.alias.html
        # Python class DataFrame with method alias(self, alias): https://github.com/apache/spark/blob/master/python/pyspark/sql/dataframe.py
        # Scala type DataFrame: https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/package.scala
        # Scala class DataSet with method as(alias: String): https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/Dataset.scala
        return obj.alias(name)
    if isinstance(obj, NDArrayWithSchema):
        result = obj
    elif isinstance(obj, np.ndarray):
        result = obj.view(NDArrayWithSchema)
    elif isinstance(obj, SeriesWithSchema):
        result = obj
    elif isinstance(obj, pd.Series):
        result = SeriesWithSchema(obj)
    elif isinstance(obj, DataFrameWithSchema):
        result = obj
    elif isinstance(obj, pd.DataFrame):
        result = DataFrameWithSchema(obj)
    elif is_list_tensor(obj):
        obj = np.array(obj)
        result = obj.view(NDArrayWithSchema)
    elif isinstance(
        obj, (pd.core.groupby.DataFrameGroupBy, pd.core.groupby.SeriesGroupBy)
    ):
        result = obj
    elif spark_installed and isinstance(obj, pyspark.sql.GroupedData):
        result = obj
    else:
        raise ValueError(f"unexpected type(obj) {type(obj)}")
    setattr(result, "table_name", name)
    return result


def get_table_name(obj):
    if spark_installed and isinstance(obj, pyspark.sql.DataFrame):
        # Python class DataFrame with field self._jdf: https://github.com/apache/spark/blob/master/python/pyspark/sql/dataframe.py
        # Scala type DataFrame: https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/package.scala
        # Scala class DataSet with field queryExecution: https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/Dataset.scala
        # Scala fields turn into Java nullary methods
        # Py4J exposes Java methods as Python methods
        # Scala class QueryExecution with field analyzed: LogicalPlan: https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/execution/QueryExecution.scala
        spark_query = obj._jdf.queryExecution().analyzed()  # type: ignore
        try:
            # calling spark_df.explain("extended") shows the analyzed contents
            # after spark_df.alias("foo"), analyzed contents should be SubqueryAlias
            # Scala class SuqueryAlias with field identifier: https://github.com/apache/spark/blob/master/sql/catalyst/src/main/scala/org/apache/spark/sql/catalyst/plans/logical/basicLogicalOperators.scala
            # str(..) converts the Java string into a Python string
            result = str(spark_query.identifier())
        except py4j.protocol.Py4JError:
            result = None
        return result
    if (
        isinstance(
            obj,
            (
                NDArrayWithSchema,
                SeriesWithSchema,
                DataFrameWithSchema,
                pd.core.groupby.DataFrameGroupBy,
                pd.core.groupby.SeriesGroupBy,
            ),
        )
        or (spark_installed and isinstance(obj, pyspark.sql.GroupedData))
    ):
        return getattr(obj, "table_name", None)
    return None


def strip_schema(obj):
    if isinstance(obj, NDArrayWithSchema):
        result = np.array(obj)
        assert type(result) == np.ndarray
    elif isinstance(obj, SeriesWithSchema):
        result = pd.Series(obj)
        assert type(result) == pd.Series
    elif isinstance(obj, DataFrameWithSchema):
        result = pd.DataFrame(obj)
        assert type(result) == pd.DataFrame
    else:
        result = obj
    return result


def dtype_to_schema(typ) -> JSON_TYPE:
    result: JSON_TYPE
    if typ is bool or np.issubdtype(typ, np.bool_):
        result = {"type": "boolean"}
    elif np.issubdtype(typ, np.unsignedinteger):
        result = {"type": "integer", "minimum": 0}
    elif np.issubdtype(typ, np.integer):
        result = {"type": "integer"}
    elif np.issubdtype(typ, np.number):
        result = {"type": "number"}
    elif np.issubdtype(typ, np.string_) or np.issubdtype(typ, np.unicode_):
        result = {"type": "string"}
    elif isinstance(typ, np.dtype):
        if typ.fields:
            props = {k: dtype_to_schema(t) for k, t in typ.fields.items()}
            result = {"type": "object", "properties": props}
        elif typ.shape:
            result = shape_and_dtype_to_schema(typ.shape, typ.subdtype)
        elif np.issubdtype(typ, np.object_):
            result = {"type": "string"}
        else:
            assert False, f"unexpected dtype {typ}"
    else:
        assert False, f"unexpected non-dtype {typ}"
    lale.type_checking.validate_is_schema(result)
    return result


def shape_and_dtype_to_schema(shape, dtype) -> JSON_TYPE:
    result = dtype_to_schema(dtype)
    for dim in reversed(shape):
        result = {"type": "array", "minItems": dim, "maxItems": dim, "items": result}
    lale.type_checking.validate_is_schema(result)
    return result


def list_tensor_to_shape_and_dtype(ls) -> Optional[Tuple[Tuple[int, ...], Type]]:
    if isinstance(ls, (int, float, str)):
        return ((), type(ls))
    if isinstance(ls, list):
        sub_result: Any = "Any"
        for item in ls:
            item_result = list_tensor_to_shape_and_dtype(item)
            if item_result is None:
                return None
            if sub_result == "Any":
                sub_result = item_result
            elif sub_result != item_result:
                return None
        if sub_result == "Any" and len(ls) == 0:
            return ((len(ls),) + (), int)
        sub_shape, sub_dtype = sub_result
        return ((len(ls),) + sub_shape, sub_dtype)
    return None


def is_list_tensor(obj) -> bool:
    if isinstance(obj, list):
        shape_and_dtype = list_tensor_to_shape_and_dtype(obj)
        return shape_and_dtype is not None
    return False


def list_tensor_to_schema(ls) -> Optional[JSON_TYPE]:
    shape_and_dtype = list_tensor_to_shape_and_dtype(ls)
    if shape_and_dtype is None:
        return None
    result = shape_and_dtype_to_schema(*shape_and_dtype)
    return result


def ndarray_to_schema(array) -> JSON_TYPE:
    assert isinstance(array, np.ndarray)
    if (
        isinstance(array, NDArrayWithSchema)
        and hasattr(array, "json_schema")
        and array.json_schema is not None
    ):
        return array.json_schema
    return shape_and_dtype_to_schema(array.shape, array.dtype)


def csr_matrix_to_schema(matrix) -> JSON_TYPE:
    assert isinstance(matrix, scipy.sparse.csr_matrix)
    result = shape_and_dtype_to_schema(matrix.shape, matrix.dtype)
    result["isSparse"] = {}  # true schema
    return result


def dataframe_to_schema(df) -> JSON_TYPE:
    assert isinstance(df, pd.DataFrame)
    if (
        isinstance(df, DataFrameWithSchema)
        and hasattr(df, "json_schema")
        and df.json_schema is not None
    ):
        return df.json_schema
    n_rows, n_columns = df.shape
    assert n_columns == len(df.columns) and n_columns == len(df.dtypes)
    items = [
        {"description": str(col), **dtype_to_schema(df.dtypes[col])}
        for col in df.columns
    ]
    result = {
        "type": "array",
        "minItems": n_rows,
        "maxItems": n_rows,
        "items": {
            "type": "array",
            "minItems": n_columns,
            "maxItems": n_columns,
            "items": items,
        },
    }
    lale.type_checking.validate_is_schema(result)
    return result


def series_to_schema(series) -> JSON_TYPE:
    assert isinstance(series, pd.Series)
    if (
        isinstance(series, SeriesWithSchema)
        and hasattr(series, "json_schema")
        and series.json_schema is not None
    ):
        return series.json_schema
    (n_rows,) = series.shape
    result = {
        "type": "array",
        "minItems": n_rows,
        "maxItems": n_rows,
        "items": {"description": str(series.name), **dtype_to_schema(series.dtype)},
    }
    lale.type_checking.validate_is_schema(result)
    return result


def torch_tensor_to_schema(tensor) -> JSON_TYPE:
    assert torch_installed, """Your Python environment does not have torch installed. You can install it with
    pip install torch
or with
    pip install 'lale[full]'"""
    assert isinstance(tensor, torch.Tensor)
    result: JSON_TYPE
    # https://pytorch.org/docs/stable/tensor_attributes.html#torch-dtype
    if tensor.dtype == torch.bool:
        result = {"type": "boolean"}
    elif tensor.dtype == torch.uint8:
        result = {"type": "integer", "minimum": 0, "maximum": 255}
    elif torch.is_floating_point(tensor):
        result = {"type": "number"}
    else:
        result = {"type": "integer"}
    for dim in reversed(tensor.shape):
        result = {"type": "array", "minItems": dim, "maxItems": dim, "items": result}
    return result


def is_liac_arff(obj) -> bool:
    expected_types = {
        "description": str,
        "relation": str,
        "attributes": list,
        "data": list,
    }
    if not isinstance(obj, dict):
        return False
    for k, t in expected_types.items():
        if k not in obj or not isinstance(obj[k], t):
            return False
    return True


def liac_arff_to_schema(larff) -> JSON_TYPE:
    assert is_liac_arff(
        larff
    ), """Your Python environment might contain an 'arff' package different from 'liac-arff'. You can install it with
    pip install 'liac-arff>=2.4.0'
or with
    pip install 'lale[full]'"""
    n_rows, n_columns = len(larff["data"]), len(larff["attributes"])

    def larff_type_to_schema(larff_type) -> JSON_TYPE:
        if isinstance(larff_type, str):
            a2j = {
                "numeric": "number",
                "real": "number",
                "integer": "integer",
                "string": "string",
            }
            return {"type": a2j[larff_type.lower()]}
        assert isinstance(larff_type, list)
        return {"enum": [*larff_type]}

    items = [
        {"description": attr[0], **larff_type_to_schema(attr[1])}
        for attr in larff["attributes"]
    ]
    result = {
        "type": "array",
        "minItems": n_rows,
        "maxItems": n_rows,
        "items": {
            "type": "array",
            "minItems": n_columns,
            "maxItems": n_columns,
            "items": items,
        },
    }
    lale.type_checking.validate_is_schema(result)
    return result


def to_schema(obj) -> JSON_TYPE:
    result = None
    if obj is None:
        result = {"enum": [None]}
    elif isinstance(obj, np.ndarray):
        result = ndarray_to_schema(obj)
    elif isinstance(obj, scipy.sparse.csr_matrix):
        result = csr_matrix_to_schema(obj)
    elif isinstance(obj, pd.DataFrame):
        result = dataframe_to_schema(obj)
    elif isinstance(obj, pd.Series):
        result = series_to_schema(obj)
    elif torch_installed and isinstance(obj, torch.Tensor):
        result = torch_tensor_to_schema(obj)
    elif is_liac_arff(obj):
        result = liac_arff_to_schema(obj)
    elif lale.type_checking.is_schema(obj):
        result = obj
    elif isinstance(obj, list):
        result = list_tensor_to_schema(obj)
    elif _is_spark_df(obj):
        result = dataframe_to_schema(obj.toPandas())
    if result is None:
        raise ValueError(f"to_schema(obj), type {type(obj)}, value {obj}")
    lale.type_checking.validate_is_schema(result)
    return result



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.discriminant_analysis

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Quadratic Discriminant Analysis",
    "allOf": [
        {
            "type": "object",
            "required": ["priors", "store_covariance"],
            "relevantToOptimizer": ["reg_param", "tol"],
            "additionalProperties": False,
            "properties": {
                "priors": {
                    "anyOf": [
                        {
                            "type": "array",
                            "items": {"type": "number"},
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Priors on classes",
                },
                "reg_param": {
                    "type": "number",
                    "minimumForOptimizer": 0.0,
                    "maximumForOptimizer": 1.0,
                    "distribution": "uniform",
                    "default": 0.0,
                    "description": "Regularizes the covariance estimate as",
                },
                "store_covariance": {
                    "type": "boolean",
                    "default": False,
                    "description": "If True the covariance matrices are computed and stored in the",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "Threshold used for rank estimation.",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "description": "Fit the model according to the given training data and parameters.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "Training vector, where n_samples is the number of samples and",
        },
        "y": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"type": "array", "items": {"type": "string"}},
                {"type": "array", "items": {"type": "boolean"}},
            ],
            "description": "Target values (integers)",
        },
    },
}
_input_predict_schema = {
    "description": "Perform classification on an array of test vectors X.",
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
        },
    },
}
_output_predict_schema = {
    "description": "Perform classification on an array of test vectors X.",
    "anyOf": [
        {"type": "array", "items": {"type": "number"}},
        {"type": "array", "items": {"type": "string"}},
        {"type": "array", "items": {"type": "boolean"}},
    ],
}

_input_predict_proba_schema = {
    "description": "Return posterior probabilities of classification.",
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "Array of samples/test vectors.",
        },
    },
}
_output_predict_proba_schema = {
    "description": "Posterior probabilities of classification per class.",
    "type": "array",
    "items": {
        "type": "array",
        "items": {"type": "number"},
    },
}

_input_decision_function_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        }
    },
}

_output_decision_function_schema = {
    "description": "Confidence scores for samples for each class in the model.",
    "anyOf": [
        {
            "description": "In the multi-way case, score per (sample, class) combination.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        {
            "description": "In the binary case, score for `self._classes[1]`.",
            "type": "array",
            "items": {"type": "number"},
        },
    ],
}

_combined_schemas = {
    "description": """`Quadratic discriminant analysis`_ classifier with a quadratic decision boundary from scikit-learn.

.. _`Quadratic discriminant analysis`: https://scikit-learn.org/stable/modules/generated/sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.quadratic_discriminant_analysis.html",
    "import_from": "sklearn.discriminant_analysis",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
        "input_decision_function": _input_decision_function_schema,
        "output_decision_function": _output_decision_function_schema,
    },
}


QuadraticDiscriminantAnalysis = lale.operators.make_operator(
    sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis, _combined_schemas
)

lale.docstrings.set_docstrings(QuadraticDiscriminantAnalysis)



from numpy import inf, nan
from sklearn.neural_network import MLPClassifier as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _MLPClassifierImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for MLPClassifier    Multi-layer Perceptron classifier.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "hidden_layer_sizes",
                "activation",
                "solver",
                "alpha",
                "batch_size",
                "learning_rate",
                "learning_rate_init",
                "power_t",
                "max_iter",
                "shuffle",
                "random_state",
                "tol",
                "verbose",
                "warm_start",
                "momentum",
                "nesterovs_momentum",
                "early_stopping",
                "validation_fraction",
                "beta_1",
                "beta_2",
                "epsilon",
                "n_iter_no_change",
            ],
            "relevantToOptimizer": [
                "activation",
                "solver",
                "alpha",
                "batch_size",
                "learning_rate",
                "max_iter",
                "shuffle",
                "tol",
                "nesterovs_momentum",
                "epsilon",
            ],
            "additionalProperties": False,
            "properties": {
                "hidden_layer_sizes": {
                    "XXX TODO XXX": "tuple, length = n_layers - 2, default (100,)",
                    "description": "The ith element represents the number of neurons in the ith hidden layer.",
                    "type": "array",
                    "laleType": "tuple",
                    "default": (100,),
                },
                "activation": {
                    "enum": ["identity", "logistic", "tanh", "relu"],
                    "default": "relu",
                    "description": "Activation function for the hidden layer",
                },
                "solver": {
                    "enum": ["lbfgs", "sgd", "adam"],
                    "default": "adam",
                    "description": "The solver for weight optimization",
                },
                "alpha": {
                    "type": "number",
                    "minimumForOptimizer": 1e-10,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "L2 penalty (regularization term) parameter.",
                },
                "batch_size": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 128,
                            "distribution": "uniform",
                        },
                        {"enum": ["auto"]},
                    ],
                    "default": "auto",
                    "description": "Size of minibatches for stochastic optimizers",
                },
                "learning_rate": {
                    "enum": ["constant", "invscaling", "adaptive"],
                    "default": "constant",
                    "description": "Learning rate schedule for weight updates",
                },
                "learning_rate_init": {
                    "type": "number",
                    "default": 0.001,
                    "description": "The initial learning rate used",
                },
                "power_t": {
                    "type": "number",
                    "default": 0.5,
                    "description": "The exponent for inverse scaling learning rate",
                },
                "max_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 200,
                    "description": "Maximum number of iterations",
                },
                "shuffle": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to shuffle samples in each iteration",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "Tolerance for the optimization",
                },
                "verbose": {
                    "type": "boolean",
                    "default": False,
                    "description": "Whether to print progress messages to stdout.",
                },
                "warm_start": {
                    "type": "boolean",
                    "default": False,
                    "description": "When set to True, reuse the solution of the previous call to fit as initialization, otherwise, just erase the previous solution",
                },
                "momentum": {
                    "type": "number",
                    "default": 0.9,
                    "description": "Momentum for gradient descent update",
                },
                "nesterovs_momentum": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to use Nesterov's momentum",
                },
                "early_stopping": {
                    "type": "boolean",
                    "default": False,
                    "description": "Whether to use early stopping to terminate training when validation score is not improving",
                },
                "validation_fraction": {
                    "type": "number",
                    "default": 0.1,
                    "description": "The proportion of training data to set aside as validation set for early stopping",
                },
                "beta_1": {
                    "type": "number",
                    "default": 0.9,
                    "description": "Exponential decay rate for estimates of first moment vector in adam, should be in [0, 1)",
                },
                "beta_2": {
                    "type": "number",
                    "default": 0.999,
                    "description": "Exponential decay rate for estimates of second moment vector in adam, should be in [0, 1)",
                },
                "epsilon": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 1.35,
                    "distribution": "loguniform",
                    "default": 1e-08,
                    "description": "Value for numerical stability in adam",
                },
                "n_iter_no_change": {
                    "type": "integer",
                    "default": 10,
                    "description": "Maximum number of epochs to not meet ``tol`` improvement",
                },
            },
        },
        {
            "description": "learning_rate, only used when solver='sgd'",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"learning_rate": {"enum": ["constant"]}},
                },
                {"type": "object", "properties": {"solver": {"enum": ["sgd"]}}},
            ],
        },
        {
            "description": "learning_rate_init, only used when solver='sgd' or 'adam'",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"learning_rate_init": {"enum": [0.001]}},
                },
                {"type": "object", "properties": {"solver": {"enum": ["sgd", "adam"]}}},
            ],
        },
        {
            "description": "power_t, only used when solver='sgd'",
            "anyOf": [
                {"type": "object", "properties": {"power_t": {"enum": [0.5]}}},
                {"type": "object", "properties": {"solver": {"enum": ["sgd"]}}},
            ],
        },
        {
            "description": "shuffle, only used when solver='sgd' or 'adam'",
            "anyOf": [
                {"type": "object", "properties": {"shuffle": {"enum": [True]}}},
                {"type": "object", "properties": {"solver": {"enum": ["sgd", "adam"]}}},
            ],
        },
        {
            "description": "momentum, only used when solver='sgd'",
            "anyOf": [
                {"type": "object", "properties": {"momentum": {"enum": [0.9]}}},
                {"type": "object", "properties": {"solver": {"enum": ["sgd"]}}},
            ],
        },
        {
            "XXX TODO XXX": "Parameter: nesterovs_momentum > only used when solver='sgd' and momentum > 0"
        },
        {
            "description": "early_stopping, only effective when solver='sgd' or 'adam'",
            "anyOf": [
                {"type": "object", "properties": {"early_stopping": {"enum": [False]}}},
                {"type": "object", "properties": {"solver": {"enum": ["sgd", "adam"]}}},
            ],
        },
        {
            "description": "validation_fraction, only used if early_stopping is true",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"validation_fraction": {"enum": [0.1]}},
                },
                {"type": "object", "properties": {"early_stopping": {"enum": [True]}}},
            ],
        },
        {
            "description": "beta_1, only used when solver='adam'",
            "anyOf": [
                {"type": "object", "properties": {"beta_1": {"enum": [0.9]}}},
                {"type": "object", "properties": {"solver": {"enum": ["adam"]}}},
            ],
        },
        {
            "description": "beta_2, only used when solver='adam'",
            "anyOf": [
                {"type": "object", "properties": {"beta_2": {"enum": [0.999]}}},
                {"type": "object", "properties": {"solver": {"enum": ["adam"]}}},
            ],
        },
        {
            "description": "epsilon, only used when solver='adam'",
            "anyOf": [
                {"type": "object", "properties": {"epsilon": {"enum": [1e-08]}}},
                {"type": "object", "properties": {"solver": {"enum": ["adam"]}}},
            ],
        },
        {
            "description": "n_iter_no_change, only effective when solver='sgd' or 'adam' ",
            "anyOf": [
                {"type": "object", "properties": {"n_iter_no_change": {"enum": [10]}}},
                {"type": "object", "properties": {"solver": {"enum": ["sgd", "adam"]}}},
            ],
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit the model to data matrix X and target(s) y.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array-like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "The input data.",
        },
        "y": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "The target values (class labels in classification, real numbers in regression).",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict using the multi-layer perceptron classifier",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "The input data.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "The predicted classes.",
    "anyOf": [
        {"type": "array", "items": {"type": "number"}},
        {"type": "array", "items": {"type": "array", "items": {"type": "number"}}},
    ],
}
_input_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Probability estimates.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "The input data.",
        }
    },
}
_output_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "The predicted probability of the sample for each class in the model, where classes are ordered as they are in `self.classes_`.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.neural_network.MLPClassifier#sklearn-neural_network-mlpclassifier",
    "import_from": "sklearn.neural_network",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}
MLPClassifier = make_operator(_MLPClassifierImpl, _combined_schemas)

set_docstrings(MLPClassifier)



from numpy import inf, nan
from sklearn.linear_model import RidgeClassifier as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _RidgeClassifierImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for RidgeClassifier    Classifier using Ridge regression.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "alpha",
                "fit_intercept",
                "normalize",
                "copy_X",
                "max_iter",
                "tol",
                "class_weight",
                "solver",
                "random_state",
            ],
            "relevantToOptimizer": [
                "alpha",
                "fit_intercept",
                "normalize",
                "copy_X",
                "max_iter",
                "tol",
                "solver",
            ],
            "additionalProperties": False,
            "properties": {
                "alpha": {
                    "type": "number",
                    "minimumForOptimizer": 1e-10,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 1.0,
                    "description": "Regularization strength; must be a positive float",
                },
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to calculate the intercept for this model",
                },
                "normalize": {
                    "type": "boolean",
                    "default": False,
                    "description": "This parameter is ignored when ``fit_intercept`` is set to False",
                },
                "copy_X": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True, X will be copied; else, it may be overwritten.",
                },
                "max_iter": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 10,
                            "maximumForOptimizer": 1000,
                            "distribution": "uniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Maximum number of iterations for conjugate gradient solver",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.001,
                    "description": "Precision of the solution.",
                },
                "class_weight": {
                    "XXX TODO XXX": "dict or 'balanced', optional",
                    "description": "Weights associated with classes in the form ``{class_label: weight}``",
                    "enum": ["balanced"],
                    "default": "balanced",
                },
                "solver": {
                    "enum": [
                        "auto",
                        "svd",
                        "cholesky",
                        "lsqr",
                        "sparse_cg",
                        "sag",
                        "saga",
                    ],
                    "default": "auto",
                    "description": "Solver to use in the computational routines:  - 'auto' chooses the solver automatically based on the type of data",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "The seed of the pseudo random number generator to use when shuffling the data",
                },
            },
        },
        {
            "XXX TODO XXX": "Parameter: solver > only guaranteed on features with approximately the same scale"
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit Ridge regression model.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training data",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values",
        },
        "sample_weight": {
            "anyOf": [
                {"type": "number"},
                {"type": "array", "items": {"type": "number"}},
            ],
            "description": "Sample weight",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict class labels for samples in X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predicted class label per sample.",
    "type": "array",
    "items": {"type": "number"},
}
_input_decision_function_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict confidence scores for samples.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_decision_function_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Confidence scores per (sample, class) combination",
    "laleType": "Any",
    "XXX TODO XXX": "array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)",
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.RidgeClassifier#sklearn-linear_model-ridgeclassifier",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_decision_function": _input_decision_function_schema,
        "output_decision_function": _output_decision_function_schema,
    },
}
RidgeClassifier = make_operator(_RidgeClassifierImpl, _combined_schemas)

set_docstrings(RidgeClassifier)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import pandas as pd
import sklearn.impute

import lale.docstrings
import lale.operators


class _SimpleImputerImpl:
    def __init__(
        self,
        missing_values=None,
        strategy="mean",
        fill_value=None,
        verbose=0,
        copy=True,
        add_indicator=False,
    ):
        self._hyperparams = {
            "missing_values": missing_values,
            "strategy": strategy,
            "fill_value": fill_value,
            "verbose": verbose,
            "copy": copy,
            "add_indicator": add_indicator,
        }
        self._wrapped_model = sklearn.impute.SimpleImputer(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        result = self._wrapped_model.transform(X)
        if isinstance(X, pd.DataFrame):
            result = pd.DataFrame(data=result, index=X.index, columns=X.columns)
        return result

    def transform_schema(self, s_X):
        return s_X


_hyperparams_schema = {
    "description": "Imputation transformer for completing missing values.",
    "allOf": [
        {
            "type": "object",
            "additionalProperties": False,
            "required": [
                "missing_values",
                "strategy",
                "fill_value",
                "verbose",
                "copy",
                "add_indicator",
            ],
            "relevantToOptimizer": ["strategy"],
            "properties": {
                "missing_values": {
                    "anyOf": [
                        {"type": "number"},
                        {"type": "string"},
                        {"enum": [np.nan]},
                        {"enum": [None]},
                    ],
                    "default": np.nan,
                    "description": "The placeholder for the missing values.",
                },
                "strategy": {
                    "anyOf": [
                        {"enum": ["constant"], "forOptimizer": False},
                        {"enum": ["mean", "median", "most_frequent"]},
                    ],
                    "default": "mean",
                    "description": "The imputation strategy.",
                },
                "fill_value": {
                    "anyOf": [{"type": "number"}, {"type": "string"}, {"enum": [None]}],
                    "default": None,
                    "description": 'When strategy == "constant", fill_value is used to replace all occurrences of missing_values',
                },
                "verbose": {
                    "type": "integer",
                    "default": 0,
                    "description": "Controls the verbosity of the imputer.",
                },
                "copy": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True, a copy of X will be created.",
                },
                "add_indicator": {
                    "type": "boolean",
                    "default": False,
                    "description": "If True, a MissingIndicator transform will stack onto output of the imputer’s transform.",
                },
            },
        },
        {
            "description": "Imputation not possible when missing_values == 0 and input is sparse. Provide a dense array instead.",
            "anyOf": [
                {"type": "object", "laleNot": "X/isSparse"},
                {
                    "type": "object",
                    "properties": {"missing_values": {"not": {"enum": [0]}}},
                },
            ],
        },
    ],
}

_input_fit_schema = {
    "description": "Fit the imputer on X.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
            "description": "Input data, where ``n_samples`` is the number of samples and  ``n_features`` is the number of features.",
        },
        "y": {},
    },
}

_input_transform_schema = {
    "description": "Impute all missing values in X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
            "description": "The input data to complete.",
        },
    },
}
_output_transform_schema = {
    "description": "The input data to complete.",
    "type": "array",
    "items": {
        "type": "array",
        "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
    },
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Simple imputer`_ transformer from scikit-learn for completing missing values.

.. _`Simple imputer`: https://scikit-learn.org/stable/modules/generated/sklearn.impute.SimpleImputer.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.simple_imputer.html",
    "import_from": "sklearn.impute",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


SimpleImputer = lale.operators.make_operator(_SimpleImputerImpl, _combined_schemas)

lale.docstrings.set_docstrings(SimpleImputer)



disable_hyperparams_schema_validation = False
disable_data_schema_validation = True


def set_disable_data_schema_validation(flag: bool):
    """Lale can validate the input and output data used for fit, predict, predict_proba etc.
    against the data schemas defined for an operator. This method allows users to control
    whether the data schema validation should be turned on or not.

    Parameters
    ----------
    flag : bool
        A value of True will disable the data schema validation, and a value of False will enable it.
        It is True by default.
    """
    global disable_data_schema_validation
    disable_data_schema_validation = flag


def set_disable_hyperparams_schema_validation(flag: bool):
    """Lale can validate the hyperparameter values passed while creating an operator against
    the json schema defined for hyperparameters of an operator. This method allows users to control
    whether such validation should be turned on or not.

    Parameters
    ----------
    flag : bool
        A value of True will disable the hyperparameter schema validation, and a value of False will enable it.
        It is False by default.
    """
    global disable_hyperparams_schema_validation
    disable_hyperparams_schema_validation = flag



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.preprocessing

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Normalize samples individually to unit norm.",
    "allOf": [
        {
            "type": "object",
            "required": ["norm"],
            "relevantToOptimizer": ["norm"],
            "additionalProperties": False,
            "properties": {
                "norm": {
                    "enum": ["l1", "l2", "max"],
                    "default": "l2",
                    "description": "The norm to use to normalize each non zero sample.",
                },
                "copy": {
                    "type": "boolean",
                    "default": True,
                    "description": "Set to False to perform inplace row normalization and avoid a copy.",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "description": "Do nothing and return the estimator unchanged",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        "y": {"description": "Target class labels; the array is over samples."},
    },
}

_input_transform_schema = {
    "description": "Scale each non zero row of X to unit norm",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The data to normalize, row by row. scipy.sparse matrices should be",
        },
        "copy": {
            "anyOf": [{"type": "boolean"}, {"enum": [None]}],
            "default": None,
            "description": "Copy the input X or not.",
        },
    },
}
_output_transform_schema = {
    "description": "Scale each non zero row of X to unit norm",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Normalizer`_ transformer from scikit-learn.

.. _`Normalizer`: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Normalizer.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.normalizer.html",
    "import_from": "sklearn.preprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


Normalizer = lale.operators.make_operator(
    sklearn.preprocessing.Normalizer, _combined_schemas
)

lale.docstrings.set_docstrings(Normalizer)
